# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.

from gosling.schemapi import SchemaBase, Undefined, _subclasses

import pkgutil
import json

def load_schema():
    """Load the json schema associated with this module's functions"""
    return json.loads(pkgutil.get_data(__name__, 'gosling-schema.json').decode('utf-8'))


class GoslingSchema(SchemaBase):
    _rootschema = load_schema()
    @classmethod
    def _default_wrapper_classes(cls):
        return _subclasses(GoslingSchema)


class Root(GoslingSchema):
    """Root schema wrapper

    anyOf(Mapping(required=[alignment, tracks]), Mapping(required=[tracks]),
    Mapping(required=[tracks]), Mapping(required=[views]))
    """
    _schema = GoslingSchema._rootschema
    _rootschema = _schema

    def __init__(self, *args, **kwds):
        super(Root, self).__init__(*args, **kwds)


class Aggregate(GoslingSchema):
    """Aggregate schema wrapper

    enum('max', 'min', 'mean', 'bin', 'count')
    """
    _schema = {'$ref': '#/definitions/Aggregate'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(Aggregate, self).__init__(*args)


class Assembly(GoslingSchema):
    """Assembly schema wrapper

    anyOf(string, string, string, string, string, string, string, string, :class:`ChromSizes`)
    """
    _schema = {'$ref': '#/definitions/Assembly'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(Assembly, self).__init__(*args, **kwds)


class AxisPosition(GoslingSchema):
    """AxisPosition schema wrapper

    enum('none', 'top', 'bottom', 'left', 'right')
    """
    _schema = {'$ref': '#/definitions/AxisPosition'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(AxisPosition, self).__init__(*args)


class BinAggregate(GoslingSchema):
    """BinAggregate schema wrapper

    enum('mean', 'sum')
    """
    _schema = {'$ref': '#/definitions/BinAggregate'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(BinAggregate, self).__init__(*args)


class Channel(GoslingSchema):
    """Channel schema wrapper

    anyOf(:class:`ChannelDeep`, :class:`ChannelValue`)
    """
    _schema = {'$ref': '#/definitions/Channel'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(Channel, self).__init__(*args, **kwds)


class ChannelDeep(Channel):
    """ChannelDeep schema wrapper

    anyOf(:class:`X`, :class:`Y`, :class:`Row`, :class:`Color`, :class:`Size`, :class:`Stroke`,
    :class:`StrokeWidth`, :class:`Opacity`, :class:`Text`)
    """
    _schema = {'$ref': '#/definitions/ChannelDeep'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(ChannelDeep, self).__init__(*args, **kwds)


class ChannelValue(Channel):
    """ChannelValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _schema = {'$ref': '#/definitions/ChannelValue'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, value=Undefined, **kwds):
        super(ChannelValue, self).__init__(value=value, **kwds)


class ChromSizes(Assembly):
    """ChromSizes schema wrapper

    List(List([string, float]))
    Custom chromosome sizes, e.g., [["foo", 1000], ["bar", 300], ["baz", 240]]
    """
    _schema = {'$ref': '#/definitions/ChromSizes'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(ChromSizes, self).__init__(*args)


class Color(ChannelDeep):
    """Color schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    legend : boolean
        Whether to display legend. __Default__: `false`
    range : :class:`Range`
        Determine the colors that should be bound to data value. Default properties are
        determined considering the field type.
    scale : enum('linear', 'log')

    scaleOffset : List(float)
        Whether to use offset of the domain proportionally. This is bound to brushes on the
        color legend. __Default__: `[0, 1]`
    title : string
        Title of the legend. __Default__: `undefined`
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/Color'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, domain=Undefined, field=Undefined, legend=Undefined, range=Undefined,
                 scale=Undefined, scaleOffset=Undefined, title=Undefined, type=Undefined, **kwds):
        super(Color, self).__init__(domain=domain, field=field, legend=legend, range=range, scale=scale,
                                    scaleOffset=scaleOffset, title=title, type=type, **kwds)


class DataDeep(GoslingSchema):
    """DataDeep schema wrapper

    anyOf(:class:`JsonData`, :class:`CsvData`, :class:`BedData`, :class:`BigWigData`,
    :class:`MultivecData`, :class:`BeddbData`, :class:`VectorData`, :class:`MatrixData`,
    :class:`BamData`, :class:`VcfData`, :class:`GffData`)
    """
    _schema = {'$ref': '#/definitions/DataDeep'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(DataDeep, self).__init__(*args, **kwds)


class BamData(DataDeep):
    """BamData schema wrapper

    Mapping(required=[type, url, indexUrl])
    Binary Alignment Map (BAM) is the comprehensive raw data of genome sequencing; it consists
    of the lossless, compressed binary representation of the Sequence Alignment Map-files.

    Attributes
    ----------

    indexUrl : string
        URL link to the index file of the BAM file
    type : string

    url : string
        URL link to the BAM data file
    extractJunction : boolean
        Determine whether to extract exon-to-exon junctions. __Default__: `false`
    junctionMinCoverage : float
        Determine the threshold of coverage when extracting exon-to-exon junctions.
        __Default__: `1`
    loadMates : boolean
        Load mates that are located in the same chromosome. __Default__: `false`
    maxInsertSize : float
        Determines the threshold of insert sizes for determining the structural variants.
        __Default__: `5000`
    """
    _schema = {'$ref': '#/definitions/BamData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, indexUrl=Undefined, type=Undefined, url=Undefined, extractJunction=Undefined,
                 junctionMinCoverage=Undefined, loadMates=Undefined, maxInsertSize=Undefined, **kwds):
        super(BamData, self).__init__(indexUrl=indexUrl, type=type, url=url,
                                      extractJunction=extractJunction,
                                      junctionMinCoverage=junctionMinCoverage, loadMates=loadMates,
                                      maxInsertSize=maxInsertSize, **kwds)


class BedData(DataDeep):
    """BedData schema wrapper

    Mapping(required=[type, url, indexUrl])
    BED file format

    Attributes
    ----------

    indexUrl : string
        Specify the URL address of the data file index.
    type : string

    url : string
        Specify the URL address of the data file.
    customFields : List(string)
        An array of strings, where each string is the name of a non-standard field in the
        BED file. If there are `n` custom fields, we assume that the last `n` columns of the
        BED file correspond to the custom fields.
    sampleLength : float
        Specify the number of rows loaded from the URL.

        __Default:__ `1000`
    """
    _schema = {'$ref': '#/definitions/BedData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, indexUrl=Undefined, type=Undefined, url=Undefined, customFields=Undefined,
                 sampleLength=Undefined, **kwds):
        super(BedData, self).__init__(indexUrl=indexUrl, type=type, url=url, customFields=customFields,
                                      sampleLength=sampleLength, **kwds)


class BeddbData(DataDeep):
    """BeddbData schema wrapper

    Mapping(required=[type, url, genomicFields])
    Regular BED or similar files can be pre-aggregated for the scalable data exploration. Find
    our more about this format at [HiGlass
    Docs](https://docs.higlass.io/data_preparation.html#bed-files).

    Attributes
    ----------

    genomicFields : List(Mapping(required=[index, name]))
        Specify the name of genomic data fields.
    type : string

    url : string
        Specify the URL address of the data file.
    exonIntervalFields : List([Mapping(required=[index, name]), Mapping(required=[index,
    name])])
        experimental
    valueFields : List(Mapping(required=[index, name, type]))
        Specify the column indexes, field names, and field types.
    """
    _schema = {'$ref': '#/definitions/BeddbData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, genomicFields=Undefined, type=Undefined, url=Undefined,
                 exonIntervalFields=Undefined, valueFields=Undefined, **kwds):
        super(BeddbData, self).__init__(genomicFields=genomicFields, type=type, url=url,
                                        exonIntervalFields=exonIntervalFields, valueFields=valueFields,
                                        **kwds)


class BigWigData(DataDeep):
    """BigWigData schema wrapper

    Mapping(required=[type, url])

    Attributes
    ----------

    type : string

    url : string
        Specify the URL address of the data file.
    aggregation : :class:`BinAggregate`
        Determine aggregation function to apply within bins. __Default__: `"mean"`
    binSize : float
        Binning the genomic interval in tiles (unit size: 256).
    column : string
        Assign a field name of the middle position of genomic intervals. __Default__:
        `"position"`
    end : string
        Assign a field name of the end position of genomic intervals. __Default__: `"end"`
    start : string
        Assign a field name of the start position of genomic intervals. __Default__:
        `"start"`
    value : string
        Assign a field name of quantitative values. __Default__: `"value"`
    """
    _schema = {'$ref': '#/definitions/BigWigData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, url=Undefined, aggregation=Undefined, binSize=Undefined,
                 column=Undefined, end=Undefined, start=Undefined, value=Undefined, **kwds):
        super(BigWigData, self).__init__(type=type, url=url, aggregation=aggregation, binSize=binSize,
                                         column=column, end=end, start=start, value=value, **kwds)


class CsvData(DataDeep):
    """CsvData schema wrapper

    Mapping(required=[type, url])
    Any small enough tabular data files, such as tsv, csv, BED, BEDPE, and GFF, can be loaded
    using "csv" data specification.

    Attributes
    ----------

    type : string

    url : string
        Specify the URL address of the data file.
    chromosomeField : string
        Specify the name of chromosome data fields.
    chromosomePrefix : string
        Specify the chromosome prefix if chromosomes are denoted using a prefix besides
        "chr" or a number
    genomicFields : List(string)
        Specify the name of genomic data fields.
    genomicFieldsToConvert : List(Mapping(required=[chromosomeField, genomicFields]))
        experimental
    headerNames : List(string)
        Specify the names of data fields if a CSV file does not contain a header.
    longToWideId : string
        experimental
    sampleLength : float
        Specify the number of rows loaded from the URL.

        __Default:__ `1000`
    separator : string
        Specify file separator, __Default:__ ','
    """
    _schema = {'$ref': '#/definitions/CsvData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, url=Undefined, chromosomeField=Undefined,
                 chromosomePrefix=Undefined, genomicFields=Undefined, genomicFieldsToConvert=Undefined,
                 headerNames=Undefined, longToWideId=Undefined, sampleLength=Undefined,
                 separator=Undefined, **kwds):
        super(CsvData, self).__init__(type=type, url=url, chromosomeField=chromosomeField,
                                      chromosomePrefix=chromosomePrefix, genomicFields=genomicFields,
                                      genomicFieldsToConvert=genomicFieldsToConvert,
                                      headerNames=headerNames, longToWideId=longToWideId,
                                      sampleLength=sampleLength, separator=separator, **kwds)


class DataTransform(GoslingSchema):
    """DataTransform schema wrapper

    anyOf(:class:`FilterTransform`, :class:`StrConcatTransform`, :class:`StrReplaceTransform`,
    :class:`LogTransform`, :class:`DisplaceTransform`, :class:`ExonSplitTransform`,
    :class:`GenomicLengthTransform`, :class:`SvTypeTransform`, :class:`CoverageTransform`,
    :class:`JsonParseTransform`)
    """
    _schema = {'$ref': '#/definitions/DataTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(DataTransform, self).__init__(*args, **kwds)


class CoverageTransform(DataTransform):
    """CoverageTransform schema wrapper

    Mapping(required=[type, startField, endField])
    Aggregate rows and calculate coverage

    Attributes
    ----------

    endField : string

    startField : string

    type : string

    groupField : string
        The name of a nominal field to group rows by in prior to piling-up
    newField : string

    """
    _schema = {'$ref': '#/definitions/CoverageTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, endField=Undefined, startField=Undefined, type=Undefined, groupField=Undefined,
                 newField=Undefined, **kwds):
        super(CoverageTransform, self).__init__(endField=endField, startField=startField, type=type,
                                                groupField=groupField, newField=newField, **kwds)


class Datum(GoslingSchema):
    """Datum schema wrapper

    Mapping(required=[])
    Values in the form of JSON.
    """
    _schema = {'$ref': '#/definitions/Datum'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, **kwds):
        super(Datum, self).__init__(**kwds)


class DisplaceTransform(DataTransform):
    """DisplaceTransform schema wrapper

    Mapping(required=[type, boundingBox, method, newField])

    Attributes
    ----------

    boundingBox : Mapping(required=[startField, endField])

    method : :class:`DisplacementType`
        A string that specifies the type of displacement.
    newField : string

    type : string

    maxRows : float
        Specify maximum rows to be generated (default has no limit).
    """
    _schema = {'$ref': '#/definitions/DisplaceTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, boundingBox=Undefined, method=Undefined, newField=Undefined, type=Undefined,
                 maxRows=Undefined, **kwds):
        super(DisplaceTransform, self).__init__(boundingBox=boundingBox, method=method,
                                                newField=newField, type=type, maxRows=maxRows, **kwds)


class Displacement(GoslingSchema):
    """Displacement schema wrapper

    Mapping(required=[type])

    Attributes
    ----------

    type : :class:`DisplacementType`

    padding : float

    """
    _schema = {'$ref': '#/definitions/Displacement'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, padding=Undefined, **kwds):
        super(Displacement, self).__init__(type=type, padding=padding, **kwds)


class DisplacementType(GoslingSchema):
    """DisplacementType schema wrapper

    enum('pile', 'spread')
    """
    _schema = {'$ref': '#/definitions/DisplacementType'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(DisplacementType, self).__init__(*args)


class DummyTrackStyle(GoslingSchema):
    """DummyTrackStyle schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    background : string
        Background color of the track
    outline : string
        Color of the outline of the track
    textFontSize : float
        Specify the font size of the title
    textFontWeight : enum('bold', 'normal')
        Specify the font weight of the title.
    textStroke : string
        Specify the stroke color of title.
    textStrokeWidth : float
        Specify the stroke width of the title.
    """
    _schema = {'$ref': '#/definitions/DummyTrackStyle'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, background=Undefined, outline=Undefined, textFontSize=Undefined,
                 textFontWeight=Undefined, textStroke=Undefined, textStrokeWidth=Undefined, **kwds):
        super(DummyTrackStyle, self).__init__(background=background, outline=outline,
                                              textFontSize=textFontSize, textFontWeight=textFontWeight,
                                              textStroke=textStroke, textStrokeWidth=textStrokeWidth,
                                              **kwds)


class EventStyle(GoslingSchema):
    """EventStyle schema wrapper

    Mapping(required=[])
    The styles defined here will be applied to the target marks of mouse events, such as a point
    mark after the user clicks on it.

    Attributes
    ----------

    arrange : enum('behind', 'front')
        Show event effects behind or in front of marks.
    color : string
        color of the marks when mouse events are triggered
    opacity : float
        opacity of the marks when mouse events are triggered
    stroke : string
        stroke color of the marks when mouse events are triggered
    strokeOpacity : float

    strokeWidth : float
        stroke width of the marks when mouse events are triggered
    """
    _schema = {'$ref': '#/definitions/EventStyle'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, arrange=Undefined, color=Undefined, opacity=Undefined, stroke=Undefined,
                 strokeOpacity=Undefined, strokeWidth=Undefined, **kwds):
        super(EventStyle, self).__init__(arrange=arrange, color=color, opacity=opacity, stroke=stroke,
                                         strokeOpacity=strokeOpacity, strokeWidth=strokeWidth, **kwds)


class ExonSplitTransform(DataTransform):
    """ExonSplitTransform schema wrapper

    Mapping(required=[type, separator, flag, fields])

    Attributes
    ----------

    fields : List(Mapping(required=[field, type, newField, chrField]))

    flag : Mapping(required=[field, value])

    separator : string

    type : string

    """
    _schema = {'$ref': '#/definitions/ExonSplitTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, fields=Undefined, flag=Undefined, separator=Undefined, type=Undefined, **kwds):
        super(ExonSplitTransform, self).__init__(fields=fields, flag=flag, separator=separator,
                                                 type=type, **kwds)


class FieldType(GoslingSchema):
    """FieldType schema wrapper

    enum('genomic', 'nominal', 'quantitative')
    """
    _schema = {'$ref': '#/definitions/FieldType'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(FieldType, self).__init__(*args)


class FilterTransform(DataTransform):
    """FilterTransform schema wrapper

    anyOf(:class:`OneOfFilter`, :class:`RangeFilter`, :class:`IncludeFilter`)
    """
    _schema = {'$ref': '#/definitions/FilterTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(FilterTransform, self).__init__(*args, **kwds)


class GenomicDomain(GoslingSchema):
    """GenomicDomain schema wrapper

    anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
    """
    _schema = {'$ref': '#/definitions/GenomicDomain'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(GenomicDomain, self).__init__(*args, **kwds)


class DomainChr(GenomicDomain):
    """DomainChr schema wrapper

    Mapping(required=[chromosome])

    Attributes
    ----------

    chromosome : string

    """
    _schema = {'$ref': '#/definitions/DomainChr'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, chromosome=Undefined, **kwds):
        super(DomainChr, self).__init__(chromosome=chromosome, **kwds)


class DomainChrInterval(GenomicDomain):
    """DomainChrInterval schema wrapper

    Mapping(required=[chromosome, interval])

    Attributes
    ----------

    chromosome : string
        If specified, only showing a certain interval in a chromosome.
    interval : List(float)

    """
    _schema = {'$ref': '#/definitions/DomainChrInterval'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, chromosome=Undefined, interval=Undefined, **kwds):
        super(DomainChrInterval, self).__init__(chromosome=chromosome, interval=interval, **kwds)


class DomainInterval(GenomicDomain):
    """DomainInterval schema wrapper

    Mapping(required=[interval])

    Attributes
    ----------

    interval : List(float)
        Show a certain interval within entire chromosome
    """
    _schema = {'$ref': '#/definitions/DomainInterval'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, interval=Undefined, **kwds):
        super(DomainInterval, self).__init__(interval=interval, **kwds)


class GenomicLengthTransform(DataTransform):
    """GenomicLengthTransform schema wrapper

    Mapping(required=[type, startField, endField, newField])
    Calculate genomic length using two genomic fields

    Attributes
    ----------

    endField : string

    newField : string

    startField : string

    type : string

    """
    _schema = {'$ref': '#/definitions/GenomicLengthTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, endField=Undefined, newField=Undefined, startField=Undefined, type=Undefined,
                 **kwds):
        super(GenomicLengthTransform, self).__init__(endField=endField, newField=newField,
                                                     startField=startField, type=type, **kwds)


class GffData(DataDeep):
    """GffData schema wrapper

    Mapping(required=[type, url, indexUrl])
    Generic Feature Format Version 3 (GFF3) format data. It parses files that follow the [GFF3
    specification](https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md).

    Attributes
    ----------

    indexUrl : string
        URL link to the tabix index file
    type : string

    url : string
        URL link to the GFF file
    attributesToFields : List(Mapping(required=[attribute, defaultValue]))
        Specifies which attributes to include as a fields. GFF files have an "attributes"
        column which contains a list of attributes which are each tag-value pairs
        (`tag=value`). This option allows for specific attributes to be accessible as a
        field. For example, if you have an attribute called "gene_name" and you want label
        features on your track using those values, you can use this option so that you can
        use `"field": "gene_name"` in the schema.

        If there is a single `value` corresponding to the `tag`, Gosling will parse that
        value as a string. If there are multiple `value`s corresponding to a `tag`, Gosling
        will parse it as a comma-separated list string. If a feature does not have a
        particular attribute, then the attribute value will be set to the `defaultValue`.
    sampleLength : float
        The maximum number of samples to be shown on the track. Samples are uniformly
        randomly selected so that this threshold is not exceeded. __Default:__ `1000`
    """
    _schema = {'$ref': '#/definitions/GffData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, indexUrl=Undefined, type=Undefined, url=Undefined, attributesToFields=Undefined,
                 sampleLength=Undefined, **kwds):
        super(GffData, self).__init__(indexUrl=indexUrl, type=type, url=url,
                                      attributesToFields=attributesToFields, sampleLength=sampleLength,
                                      **kwds)


class GoslingSpec(GoslingSchema):
    """GoslingSpec schema wrapper

    anyOf(Mapping(required=[alignment, tracks]), Mapping(required=[tracks]),
    Mapping(required=[tracks]), Mapping(required=[views]))
    """
    _schema = {'$ref': '#/definitions/GoslingSpec'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(GoslingSpec, self).__init__(*args, **kwds)


class IncludeFilter(FilterTransform):
    """IncludeFilter schema wrapper

    Mapping(required=[field, include, type])

    Attributes
    ----------

    field : string
        A filter is applied based on the values of the specified data field
    include : string
        Check whether the value includes a substring.
    type : string

    not : boolean
        when `{"not": true}`, apply a NOT logical operation to the filter.

        __Default:__ `false`
    """
    _schema = {'$ref': '#/definitions/IncludeFilter'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, field=Undefined, include=Undefined, type=Undefined, **kwds):
        super(IncludeFilter, self).__init__(field=field, include=include, type=type, **kwds)


class JsonData(DataDeep):
    """JsonData schema wrapper

    Mapping(required=[type, values])
    The JSON data format allows users to include data directly in the Gosling's JSON
    specification.

    Attributes
    ----------

    type : string
        Define data type.
    values : List(:class:`Datum`)
        Values in the form of JSON.
    chromosomeField : string
        Specify the name of chromosome data fields.
    genomicFields : List(string)
        Specify the name of genomic data fields.
    genomicFieldsToConvert : List(Mapping(required=[chromosomeField, genomicFields]))
        experimental
    sampleLength : float
        Specify the number of rows loaded from the URL.

        __Default:__ `1000`
    """
    _schema = {'$ref': '#/definitions/JsonData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, values=Undefined, chromosomeField=Undefined,
                 genomicFields=Undefined, genomicFieldsToConvert=Undefined, sampleLength=Undefined,
                 **kwds):
        super(JsonData, self).__init__(type=type, values=values, chromosomeField=chromosomeField,
                                       genomicFields=genomicFields,
                                       genomicFieldsToConvert=genomicFieldsToConvert,
                                       sampleLength=sampleLength, **kwds)


class JsonParseTransform(DataTransform):
    """JsonParseTransform schema wrapper

    Mapping(required=[type, field, baseGenomicField, genomicField, genomicLengthField])
    Parse JSON Object Array and append vertically

    Attributes
    ----------

    baseGenomicField : string
        Base genomic position when parsing relative position.
    field : string
        The field that contains the JSON object array.
    genomicField : string
        Relative genomic position to parse.
    genomicLengthField : string
        Length of genomic interval.
    type : string

    """
    _schema = {'$ref': '#/definitions/JsonParseTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, baseGenomicField=Undefined, field=Undefined, genomicField=Undefined,
                 genomicLengthField=Undefined, type=Undefined, **kwds):
        super(JsonParseTransform, self).__init__(baseGenomicField=baseGenomicField, field=field,
                                                 genomicField=genomicField,
                                                 genomicLengthField=genomicLengthField, type=type,
                                                 **kwds)


class Layout(GoslingSchema):
    """Layout schema wrapper

    enum('linear', 'circular')
    """
    _schema = {'$ref': '#/definitions/Layout'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(Layout, self).__init__(*args)


class LogBase(GoslingSchema):
    """LogBase schema wrapper

    anyOf(float, string)
    """
    _schema = {'$ref': '#/definitions/LogBase'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(LogBase, self).__init__(*args, **kwds)


class LogTransform(DataTransform):
    """LogTransform schema wrapper

    Mapping(required=[type, field])

    Attributes
    ----------

    field : string

    type : string

    base : :class:`LogBase`
        If not specified, 10 is used.
    newField : string
        If specified, store transformed values in a new field.
    """
    _schema = {'$ref': '#/definitions/LogTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, field=Undefined, type=Undefined, base=Undefined, newField=Undefined, **kwds):
        super(LogTransform, self).__init__(field=field, type=type, base=base, newField=newField, **kwds)


class LogicalOperation(GoslingSchema):
    """LogicalOperation schema wrapper

    enum('less-than', 'lt', 'LT', 'greater-than', 'gt', 'GT', 'less-than-or-equal-to', 'ltet',
    'LTET', 'greater-than-or-equal-to', 'gtet', 'GTET')
    """
    _schema = {'$ref': '#/definitions/LogicalOperation'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(LogicalOperation, self).__init__(*args)


class Mark(GoslingSchema):
    """Mark schema wrapper

    enum('point', 'line', 'area', 'bar', 'rect', 'text', 'withinLink', 'betweenLink', 'rule',
    'triangleLeft', 'triangleRight', 'triangleBottom', 'brush', 'header')
    """
    _schema = {'$ref': '#/definitions/Mark'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(Mark, self).__init__(*args)


class MatrixData(DataDeep):
    """MatrixData schema wrapper

    Mapping(required=[type, url])

    Attributes
    ----------

    type : string

    url : string
        URL link to the matrix data file
    binSize : float
        Determine the number of nearby cells to aggregate. __Default__: `1`
    column : string
        The name of the first genomic field. __Default__: `x`
    row : string
        The name of the first genomic field. __Default__: `y`
    value : string
        The name of the value field. __Default__: `value`
    """
    _schema = {'$ref': '#/definitions/MatrixData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, url=Undefined, binSize=Undefined, column=Undefined,
                 row=Undefined, value=Undefined, **kwds):
        super(MatrixData, self).__init__(type=type, url=url, binSize=binSize, column=column, row=row,
                                         value=value, **kwds)


class MouseEventsDeep(GoslingSchema):
    """MouseEventsDeep schema wrapper

    Mapping(required=[])
    Options for determining mouse events in detail, e.g., turning on specific events only

    Attributes
    ----------

    click : boolean
        Whether to enable click events.
    enableMouseOverOnMultipleMarks : boolean
        Determine whether all marks underneath the mouse point should be affected by mouse
        over. __Default__: `false`
    groupMarksByField : string
        Group marks using keys in a data field. This affects how a set of marks are
        highlighted/selected by interaction. __Default__: `undefined`
    mouseOver : boolean
        Whether to enable mouseover events.
    rangeSelect : boolean
        Whether to send range selection events.
    """
    _schema = {'$ref': '#/definitions/MouseEventsDeep'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, click=Undefined, enableMouseOverOnMultipleMarks=Undefined,
                 groupMarksByField=Undefined, mouseOver=Undefined, rangeSelect=Undefined, **kwds):
        super(MouseEventsDeep, self).__init__(click=click,
                                              enableMouseOverOnMultipleMarks=enableMouseOverOnMultipleMarks,
                                              groupMarksByField=groupMarksByField, mouseOver=mouseOver,
                                              rangeSelect=rangeSelect, **kwds)


class MultipleViews(GoslingSchema):
    """MultipleViews schema wrapper

    Mapping(required=[views])

    Attributes
    ----------

    views : List(anyOf(:class:`SingleView`, :class:`MultipleViews`))
        An array of view specifications
    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    arrangement : enum('parallel', 'serial', 'horizontal', 'vertical')
        Specify how multiple views are arranged.
    assembly : :class:`Assembly`
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    id : string
        The ID of a view that is maintained for the use of JS API functions, e.g., positions
        of a view
    layout : :class:`Layout`
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    orientation : :class:`Orientation`
        Specify the orientation.
    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    style : :class:`Style`
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    zoomLimits : :class:`ZoomLimits`

    """
    _schema = {'$ref': '#/definitions/MultipleViews'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, views=Undefined, _assignedHeight=Undefined, _assignedWidth=Undefined,
                 arrangement=Undefined, assembly=Undefined, centerRadius=Undefined, id=Undefined,
                 layout=Undefined, linkingId=Undefined, orientation=Undefined, spacing=Undefined,
                 static=Undefined, style=Undefined, xAxis=Undefined, xDomain=Undefined,
                 xOffset=Undefined, yDomain=Undefined, yOffset=Undefined, zoomLimits=Undefined, **kwds):
        super(MultipleViews, self).__init__(views=views, _assignedHeight=_assignedHeight,
                                            _assignedWidth=_assignedWidth, arrangement=arrangement,
                                            assembly=assembly, centerRadius=centerRadius, id=id,
                                            layout=layout, linkingId=linkingId, orientation=orientation,
                                            spacing=spacing, static=static, style=style, xAxis=xAxis,
                                            xDomain=xDomain, xOffset=xOffset, yDomain=yDomain,
                                            yOffset=yOffset, zoomLimits=zoomLimits, **kwds)


class MultivecData(DataDeep):
    """MultivecData schema wrapper

    Mapping(required=[type, url])
    Two-dimensional quantitative values, one axis for genomic coordinate and the other for
    different samples, can be converted into HiGlass' `"multivec"` data. For example, multiple
    BigWig files can be converted into a single multivec file. You can also convert sequence
    data (FASTA) into this format where rows will be different nucleotide bases (e.g., A, T, G,
    C) and quantitative values represent the frequency. Find out more about this format at
    [HiGlass Docs](https://docs.higlass.io/data_preparation.html#multivec-files).

    Attributes
    ----------

    type : string

    url : string
        Specify the URL address of the data file.
    aggregation : :class:`BinAggregate`
        Determine aggregation function to apply within bins. __Default__: `"mean"`
    binSize : float
        Binning the genomic interval in tiles (unit size: 256).
    categories : List(string)
        assign names of individual samples.
    column : string
        Assign a field name of the middle position of genomic intervals. __Default__:
        `"position"`
    end : string
        Assign a field name of the end position of genomic intervals. __Default__: `"end"`
    row : string
        Assign a field name of samples. __Default__: `"category"`
    start : string
        Assign a field name of the start position of genomic intervals. __Default__:
        `"start"`
    value : string
        Assign a field name of quantitative values. __Default__: `"value"`
    """
    _schema = {'$ref': '#/definitions/MultivecData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, url=Undefined, aggregation=Undefined, binSize=Undefined,
                 categories=Undefined, column=Undefined, end=Undefined, row=Undefined, start=Undefined,
                 value=Undefined, **kwds):
        super(MultivecData, self).__init__(type=type, url=url, aggregation=aggregation, binSize=binSize,
                                           categories=categories, column=column, end=end, row=row,
                                           start=start, value=value, **kwds)


class OneOfFilter(FilterTransform):
    """OneOfFilter schema wrapper

    Mapping(required=[field, oneOf, type])

    Attributes
    ----------

    field : string
        A filter is applied based on the values of the specified data field
    oneOf : List(anyOf(string, float, None))
        Check whether the value is an element in the provided list.
    type : string

    not : boolean
        when `{"not": true}`, apply a NOT logical operation to the filter.

        __Default:__ `false`
    """
    _schema = {'$ref': '#/definitions/OneOfFilter'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, field=Undefined, oneOf=Undefined, type=Undefined, **kwds):
        super(OneOfFilter, self).__init__(field=field, oneOf=oneOf, type=type, **kwds)


class Opacity(ChannelDeep):
    """Opacity schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    range : :class:`ValueExtent`
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/Opacity'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, domain=Undefined, field=Undefined, range=Undefined, type=Undefined, **kwds):
        super(Opacity, self).__init__(domain=domain, field=field, range=range, type=type, **kwds)


class Orientation(GoslingSchema):
    """Orientation schema wrapper

    enum('horizontal', 'vertical')
    """
    _schema = {'$ref': '#/definitions/Orientation'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(Orientation, self).__init__(*args)


class OverlaidTracks(GoslingSchema):
    """OverlaidTracks schema wrapper

    Mapping(required=[alignment, tracks])

    Attributes
    ----------

    alignment : string

    tracks : List(:class:`PartialTrack`)

    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    _invalidTrack : boolean
        internal
    _renderingId : string
        internal
    assembly : :class:`Assembly`
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    baselineY : float

    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    color : anyOf(:class:`Color`, :class:`ChannelValue`)

    data : :class:`DataDeep`

    dataTransform : List(:class:`DataTransform`)

    displacement : :class:`Displacement`

    endAngle : float
        Specify the end angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    experimental : Mapping(required=[])

    flipY : boolean

    height : float
        Specify the track height in pixels.
    id : string
        The ID of a view that is maintained for the use of JS API functions, e.g., positions
        of a view
    innerRadius : float
        Specify the inner radius of tracks when (`{"layout": "circular"}`).
    layout : :class:`Layout`
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    mark : :class:`Mark`

    opacity : anyOf(:class:`Opacity`, :class:`ChannelValue`)

    orientation : :class:`Orientation`
        Specify the orientation.
    outerRadius : float
        Specify the outer radius of tracks when `{"layout": "circular"}`.
    overlayOnPreviousTrack : boolean

    overrideTemplate : boolean

    prerelease : Mapping(required=[])
        internal
    row : anyOf(:class:`Row`, :class:`ChannelValue`)

    size : anyOf(:class:`Size`, :class:`ChannelValue`)

    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    startAngle : float
        Specify the start angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    stretch : boolean

    stroke : anyOf(:class:`Stroke`, :class:`ChannelValue`)

    strokeWidth : anyOf(:class:`StrokeWidth`, :class:`ChannelValue`)

    style : :class:`Style`
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    subtitle : string

    text : anyOf(:class:`Text`, :class:`ChannelValue`)

    title : string
        If defined, will show the textual label on the left-top corner of a track.
    tooltip : List(:class:`Tooltip`)

    visibility : List(:class:`VisibilityCondition`)

    width : float
        Specify the track width in pixels.
    x : anyOf(:class:`X`, :class:`ChannelValue`)

    x1 : anyOf(:class:`X`, :class:`ChannelValue`)

    x1e : anyOf(:class:`X`, :class:`ChannelValue`)

    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    xe : anyOf(:class:`X`, :class:`ChannelValue`)

    y : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1 : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1e : anyOf(:class:`Y`, :class:`ChannelValue`)

    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    ye : anyOf(:class:`Y`, :class:`ChannelValue`)

    zoomLimits : :class:`ZoomLimits`

    """
    _schema = {'$ref': '#/definitions/OverlaidTracks'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, alignment=Undefined, tracks=Undefined, _assignedHeight=Undefined,
                 _assignedWidth=Undefined, _invalidTrack=Undefined, _renderingId=Undefined,
                 assembly=Undefined, baselineY=Undefined, centerRadius=Undefined, color=Undefined,
                 data=Undefined, dataTransform=Undefined, displacement=Undefined, endAngle=Undefined,
                 experimental=Undefined, flipY=Undefined, height=Undefined, id=Undefined,
                 innerRadius=Undefined, layout=Undefined, linkingId=Undefined, mark=Undefined,
                 opacity=Undefined, orientation=Undefined, outerRadius=Undefined,
                 overlayOnPreviousTrack=Undefined, overrideTemplate=Undefined, prerelease=Undefined,
                 row=Undefined, size=Undefined, spacing=Undefined, startAngle=Undefined,
                 static=Undefined, stretch=Undefined, stroke=Undefined, strokeWidth=Undefined,
                 style=Undefined, subtitle=Undefined, text=Undefined, title=Undefined,
                 tooltip=Undefined, visibility=Undefined, width=Undefined, x=Undefined, x1=Undefined,
                 x1e=Undefined, xAxis=Undefined, xDomain=Undefined, xOffset=Undefined, xe=Undefined,
                 y=Undefined, y1=Undefined, y1e=Undefined, yDomain=Undefined, yOffset=Undefined,
                 ye=Undefined, zoomLimits=Undefined, **kwds):
        super(OverlaidTracks, self).__init__(alignment=alignment, tracks=tracks,
                                             _assignedHeight=_assignedHeight,
                                             _assignedWidth=_assignedWidth, _invalidTrack=_invalidTrack,
                                             _renderingId=_renderingId, assembly=assembly,
                                             baselineY=baselineY, centerRadius=centerRadius,
                                             color=color, data=data, dataTransform=dataTransform,
                                             displacement=displacement, endAngle=endAngle,
                                             experimental=experimental, flipY=flipY, height=height,
                                             id=id, innerRadius=innerRadius, layout=layout,
                                             linkingId=linkingId, mark=mark, opacity=opacity,
                                             orientation=orientation, outerRadius=outerRadius,
                                             overlayOnPreviousTrack=overlayOnPreviousTrack,
                                             overrideTemplate=overrideTemplate, prerelease=prerelease,
                                             row=row, size=size, spacing=spacing, startAngle=startAngle,
                                             static=static, stretch=stretch, stroke=stroke,
                                             strokeWidth=strokeWidth, style=style, subtitle=subtitle,
                                             text=text, title=title, tooltip=tooltip,
                                             visibility=visibility, width=width, x=x, x1=x1, x1e=x1e,
                                             xAxis=xAxis, xDomain=xDomain, xOffset=xOffset, xe=xe, y=y,
                                             y1=y1, y1e=y1e, yDomain=yDomain, yOffset=yOffset, ye=ye,
                                             zoomLimits=zoomLimits, **kwds)


class PartialTrack(GoslingSchema):
    """PartialTrack schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    _invalidTrack : boolean
        internal
    _overlay : List(Mapping(required=[]))

    _renderingId : string
        internal
    assembly : anyOf(:class:`Assembly`, string)
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    baselineY : float

    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    color : anyOf(:class:`Color`, :class:`ChannelValue`)

    data : :class:`DataDeep`

    dataTransform : List(:class:`DataTransform`)

    displacement : :class:`Displacement`

    encoding : Mapping(required=[])

    endAngle : float
        Specify the end angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    experimental : Mapping(required=[])

    flipY : boolean

    height : float
        Specify the track height in pixels.
    id : string
        Assigned to `uid` in a HiGlass view config, used for API and caching.
    innerRadius : float
        Specify the inner radius of tracks when (`{"layout": "circular"}`).
    layout : anyOf(:class:`Layout`, string)
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    mark : :class:`Mark`

    opacity : anyOf(:class:`Opacity`, :class:`ChannelValue`)

    orientation : :class:`Orientation`
        Specify the orientation.
    outerRadius : float
        Specify the outer radius of tracks when `{"layout": "circular"}`.
    overlayOnPreviousTrack : boolean

    overrideTemplate : boolean

    prerelease : Mapping(required=[])
        internal
    row : anyOf(:class:`Row`, :class:`ChannelValue`)

    size : anyOf(:class:`Size`, :class:`ChannelValue`)

    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    startAngle : float
        Specify the start angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    static : anyOf(boolean, boolean)
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    stretch : boolean

    stroke : anyOf(:class:`Stroke`, :class:`ChannelValue`)

    strokeWidth : anyOf(:class:`StrokeWidth`, :class:`ChannelValue`)

    style : anyOf(:class:`Style`, :class:`DummyTrackStyle`)
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    subtitle : string

    template : string

    text : anyOf(:class:`Text`, :class:`ChannelValue`)

    title : string
        If defined, will show the textual label on the left-top corner of a track.
    tooltip : List(:class:`Tooltip`)

    type : string
        Used to specify the dummy track
    visibility : List(:class:`VisibilityCondition`)

    width : float
        Specify the track width in pixels.
    x : anyOf(:class:`X`, :class:`ChannelValue`)

    x1 : anyOf(:class:`X`, :class:`ChannelValue`)

    x1e : anyOf(:class:`X`, :class:`ChannelValue`)

    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    xe : anyOf(:class:`X`, :class:`ChannelValue`)

    y : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1 : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1e : anyOf(:class:`Y`, :class:`ChannelValue`)

    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    ye : anyOf(:class:`Y`, :class:`ChannelValue`)

    zoomLimits : anyOf(:class:`ZoomLimits`, List(None))

    """
    _schema = {'$ref': '#/definitions/PartialTrack'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, _assignedHeight=Undefined, _assignedWidth=Undefined, _invalidTrack=Undefined,
                 _overlay=Undefined, _renderingId=Undefined, assembly=Undefined, baselineY=Undefined,
                 centerRadius=Undefined, color=Undefined, data=Undefined, dataTransform=Undefined,
                 displacement=Undefined, encoding=Undefined, endAngle=Undefined, experimental=Undefined,
                 flipY=Undefined, height=Undefined, id=Undefined, innerRadius=Undefined,
                 layout=Undefined, linkingId=Undefined, mark=Undefined, opacity=Undefined,
                 orientation=Undefined, outerRadius=Undefined, overlayOnPreviousTrack=Undefined,
                 overrideTemplate=Undefined, prerelease=Undefined, row=Undefined, size=Undefined,
                 spacing=Undefined, startAngle=Undefined, static=Undefined, stretch=Undefined,
                 stroke=Undefined, strokeWidth=Undefined, style=Undefined, subtitle=Undefined,
                 template=Undefined, text=Undefined, title=Undefined, tooltip=Undefined, type=Undefined,
                 visibility=Undefined, width=Undefined, x=Undefined, x1=Undefined, x1e=Undefined,
                 xAxis=Undefined, xDomain=Undefined, xOffset=Undefined, xe=Undefined, y=Undefined,
                 y1=Undefined, y1e=Undefined, yDomain=Undefined, yOffset=Undefined, ye=Undefined,
                 zoomLimits=Undefined, **kwds):
        super(PartialTrack, self).__init__(_assignedHeight=_assignedHeight,
                                           _assignedWidth=_assignedWidth, _invalidTrack=_invalidTrack,
                                           _overlay=_overlay, _renderingId=_renderingId,
                                           assembly=assembly, baselineY=baselineY,
                                           centerRadius=centerRadius, color=color, data=data,
                                           dataTransform=dataTransform, displacement=displacement,
                                           encoding=encoding, endAngle=endAngle,
                                           experimental=experimental, flipY=flipY, height=height, id=id,
                                           innerRadius=innerRadius, layout=layout, linkingId=linkingId,
                                           mark=mark, opacity=opacity, orientation=orientation,
                                           outerRadius=outerRadius,
                                           overlayOnPreviousTrack=overlayOnPreviousTrack,
                                           overrideTemplate=overrideTemplate, prerelease=prerelease,
                                           row=row, size=size, spacing=spacing, startAngle=startAngle,
                                           static=static, stretch=stretch, stroke=stroke,
                                           strokeWidth=strokeWidth, style=style, subtitle=subtitle,
                                           template=template, text=text, title=title, tooltip=tooltip,
                                           type=type, visibility=visibility, width=width, x=x, x1=x1,
                                           x1e=x1e, xAxis=xAxis, xDomain=xDomain, xOffset=xOffset,
                                           xe=xe, y=y, y1=y1, y1e=y1e, yDomain=yDomain, yOffset=yOffset,
                                           ye=ye, zoomLimits=zoomLimits, **kwds)


class Range(GoslingSchema):
    """Range schema wrapper

    anyOf(:class:`ValueExtent`, :class:`PredefinedColors`)
    """
    _schema = {'$ref': '#/definitions/Range'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(Range, self).__init__(*args, **kwds)


class PredefinedColors(Range):
    """PredefinedColors schema wrapper

    enum('viridis', 'grey', 'spectral', 'warm', 'cividis', 'bupu', 'rdbu', 'hot', 'pink')
    """
    _schema = {'$ref': '#/definitions/PredefinedColors'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(PredefinedColors, self).__init__(*args)


class RangeFilter(FilterTransform):
    """RangeFilter schema wrapper

    Mapping(required=[field, inRange, type])

    Attributes
    ----------

    field : string
        A filter is applied based on the values of the specified data field
    inRange : List(float)
        Check whether the value is in a number range.
    type : string

    not : boolean
        when `{"not": true}`, apply a NOT logical operation to the filter.

        __Default:__ `false`
    """
    _schema = {'$ref': '#/definitions/RangeFilter'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, field=Undefined, inRange=Undefined, type=Undefined, **kwds):
        super(RangeFilter, self).__init__(field=field, inRange=inRange, type=type, **kwds)


class ResponsiveSize(GoslingSchema):
    """ResponsiveSize schema wrapper

    anyOf(boolean, Mapping(required=[]))
    """
    _schema = {'$ref': '#/definitions/ResponsiveSize'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(ResponsiveSize, self).__init__(*args, **kwds)


class Row(ChannelDeep):
    """Row schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    clip : boolean
        Clip row when the actual y value exceeds the max value of the y scale. Used only for
        bar marks at the moment. __Default__: `true`
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    padding : float
        Determines the size of inner white spaces on the top and bottom of individiual rows.
        __Default__: `0`
    range : :class:`ValueExtent`
        Determine the start and end position of rendering area of this track along vertical
        axis. __Default__: `[0, height]`
    type : string
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/Row'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, clip=Undefined, domain=Undefined, field=Undefined, grid=Undefined,
                 legend=Undefined, padding=Undefined, range=Undefined, type=Undefined, **kwds):
        super(Row, self).__init__(clip=clip, domain=domain, field=field, grid=grid, legend=legend,
                                  padding=padding, range=range, type=type, **kwds)


class SelectivityCondition(GoslingSchema):
    """SelectivityCondition schema wrapper

    Mapping(required=[operation, measure, threshold])

    Attributes
    ----------

    measure : enum('width', 'height', 'aspectRatio')

    operation : :class:`LogicalOperation`

    threshold : float
        Threshold in the unit of pixels.
    target : enum('self', 'container')
        Does the condition applied to the visualization itself or its container?
        __Default__: `'self'`
    """
    _schema = {'$ref': '#/definitions/SelectivityCondition'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, measure=Undefined, operation=Undefined, threshold=Undefined, target=Undefined,
                 **kwds):
        super(SelectivityCondition, self).__init__(measure=measure, operation=operation,
                                                   threshold=threshold, target=target, **kwds)


class SingleView(GoslingSchema):
    """SingleView schema wrapper

    anyOf(Mapping(required=[alignment, tracks]), Mapping(required=[tracks]),
    Mapping(required=[tracks]))
    """
    _schema = {'$ref': '#/definitions/SingleView'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(SingleView, self).__init__(*args, **kwds)


class Size(ChannelDeep):
    """Size schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    legend : boolean
        not supported: Whether to display legend. __Default__: `false`
    range : :class:`ValueExtent`
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/Size'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, domain=Undefined, field=Undefined, legend=Undefined, range=Undefined,
                 type=Undefined, **kwds):
        super(Size, self).__init__(domain=domain, field=field, legend=legend, range=range, type=type,
                                   **kwds)


class StrConcatTransform(DataTransform):
    """StrConcatTransform schema wrapper

    Mapping(required=[type, fields, newField, separator])

    Attributes
    ----------

    fields : List(string)

    newField : string

    separator : string

    type : string

    """
    _schema = {'$ref': '#/definitions/StrConcatTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, fields=Undefined, newField=Undefined, separator=Undefined, type=Undefined, **kwds):
        super(StrConcatTransform, self).__init__(fields=fields, newField=newField, separator=separator,
                                                 type=type, **kwds)


class StrReplaceTransform(DataTransform):
    """StrReplaceTransform schema wrapper

    Mapping(required=[type, field, newField, replace])

    Attributes
    ----------

    field : string

    newField : string

    replace : List(Mapping(required=[from, to]))

    type : string

    """
    _schema = {'$ref': '#/definitions/StrReplaceTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, field=Undefined, newField=Undefined, replace=Undefined, type=Undefined, **kwds):
        super(StrReplaceTransform, self).__init__(field=field, newField=newField, replace=replace,
                                                  type=type, **kwds)


class Stroke(ChannelDeep):
    """Stroke schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    legend : boolean
        Whether to display legend. __Default__: `false`
    range : :class:`Range`
        Ranges of visual channel values
    scaleOffset : List(float)
        Whether to use offset of the domain proportionally. This is bound to brushes on the
        color legend. __Default__: `[0, 1]`
    title : string
        Title of the legend. __Default__: `undefined`
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/Stroke'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, domain=Undefined, field=Undefined, legend=Undefined, range=Undefined,
                 scaleOffset=Undefined, title=Undefined, type=Undefined, **kwds):
        super(Stroke, self).__init__(domain=domain, field=field, legend=legend, range=range,
                                     scaleOffset=scaleOffset, title=title, type=type, **kwds)


class StrokeWidth(ChannelDeep):
    """StrokeWidth schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    range : :class:`ValueExtent`
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/StrokeWidth'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, domain=Undefined, field=Undefined, range=Undefined, type=Undefined, **kwds):
        super(StrokeWidth, self).__init__(domain=domain, field=field, range=range, type=type, **kwds)


class Style(GoslingSchema):
    """Style schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    align : enum('left', 'right')
        Specify the alignment of marks. This property is currently only supported for
        `triangle` marks.
    background : string

    backgroundOpacity : float

    brush : Mapping(required=[])
        Customize the style of the brush mark in the `rangeSelect` mouse event.
    curve : enum('top', 'bottom', 'left', 'right')
        Specify the curve of `rule` marks.
    dashed : List(float)
        Specify the pattern of dashes and gaps for `rule` marks.
    dx : float
        Offset the position of marks in x direction. This property is currently only
        supported for `text` marks
    dy : float
        Offset the position of marks in y direction. This property is currently only
        supported for `text` marks.
    enableSmoothPath : boolean
        Whether to enable smooth paths when drawing curves.

        __Default__: `false`
    inlineLegend : boolean
        Specify whether to show legend in a single horizontal line?
    legendTitle : string
        If defined, show legend title on the top or left
    linePattern : Mapping(required=[type, size])
        Specify the pattern of dashes and gaps for `rule` marks.
    linkConnectionType : enum('straight', 'curve', 'corner')
        Specify the connetion type of `betweenLink` marks.

        __Default__: `"corner"`
    linkMinHeight : float
        The minimum height of `withinLink` and `betweenLink` marks. Unit is a percentagle.
        __Default__: `0.5`
    linkStyle : enum('elliptical', 'circular', 'straight')
        The style of `withinLink` and `betweenLink` marks. __Default__: `'circular'`
        `'elliptical'` will be used as a default option.
    matrixExtent : enum('full', 'upper-right', 'lower-left')
        Determine to show only one side of the diagonal in a HiGlass matrix. __Default__:
        `"full"`
    mouseOver : :class:`EventStyle`
        Customize visual effects of `mouseOver` events on marks.
    outline : string

    outlineWidth : float

    select : :class:`EventStyle`
        Customize visual effects of `rangeSelect` events on marks .
    textAnchor : enum('start', 'middle', 'end')
        Specify the alignment of `text` marks to a given point.
    textFontSize : float
        Specify the font size of `text` marks. Can also be specified using the `size`
        channel option of `text` marks.
    textFontWeight : enum('bold', 'normal')
        Specify the font weight of `text` marks.
    textStroke : string
        Specify the stroke of `text` marks. Can also be specified using the `stroke` channel
        option of `text` marks.
    textStrokeWidth : float
        Specify the stroke width of `text` marks. Can also be specified using the
        `strokeWidth` channel option of `text` marks.
    withinLinkVerticalLines : boolean
        Whether to show vertical lines that connect to the baseline (axis) when `y` and `ye`
        are both used. __Default__: `false`
    """
    _schema = {'$ref': '#/definitions/Style'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                 brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                 enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                 linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                 linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                 outlineWidth=Undefined, select=Undefined, textAnchor=Undefined, textFontSize=Undefined,
                 textFontWeight=Undefined, textStroke=Undefined, textStrokeWidth=Undefined,
                 withinLinkVerticalLines=Undefined, **kwds):
        super(Style, self).__init__(align=align, background=background,
                                    backgroundOpacity=backgroundOpacity, brush=brush, curve=curve,
                                    dashed=dashed, dx=dx, dy=dy, enableSmoothPath=enableSmoothPath,
                                    inlineLegend=inlineLegend, legendTitle=legendTitle,
                                    linePattern=linePattern, linkConnectionType=linkConnectionType,
                                    linkMinHeight=linkMinHeight, linkStyle=linkStyle,
                                    matrixExtent=matrixExtent, mouseOver=mouseOver, outline=outline,
                                    outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                                    textFontSize=textFontSize, textFontWeight=textFontWeight,
                                    textStroke=textStroke, textStrokeWidth=textStrokeWidth,
                                    withinLinkVerticalLines=withinLinkVerticalLines, **kwds)


class SvTypeTransform(DataTransform):
    """SvTypeTransform schema wrapper

    Mapping(required=[type, firstBp, secondBp, newField])

    Attributes
    ----------

    firstBp : Mapping(required=[chrField, posField, strandField])
        Based on the BEDPE, infer SV types. SV types are specified as one of the following
        strings: DUP, TRA, DEL, t2tINV, h2hINV.
    newField : string

    secondBp : Mapping(required=[chrField, posField, strandField])
        Based on the BEDPE, infer SV types. SV types are specified as one of the following
        strings: DUP, TRA, DEL, t2tINV, h2hINV.
    type : string

    """
    _schema = {'$ref': '#/definitions/SvTypeTransform'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, firstBp=Undefined, newField=Undefined, secondBp=Undefined, type=Undefined, **kwds):
        super(SvTypeTransform, self).__init__(firstBp=firstBp, newField=newField, secondBp=secondBp,
                                              type=type, **kwds)


class Text(ChannelDeep):
    """Text schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    domain : List(string)
        Values of the data
    field : string
        Name of the data field
    range : List(string)
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _schema = {'$ref': '#/definitions/Text'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, domain=Undefined, field=Undefined, range=Undefined, type=Undefined, **kwds):
        super(Text, self).__init__(domain=domain, field=field, range=range, type=type, **kwds)


class Tooltip(GoslingSchema):
    """Tooltip schema wrapper

    Mapping(required=[field, type])

    Attributes
    ----------

    field : string
        Specifiy a data field whose value will show in the tooltip.
    type : :class:`FieldType`
        Type of the data field.
    alt : string
        Name of the data field for showing in the tooltip. Will use the field name if not
        specified.
    format : string
        format of the data value.
    """
    _schema = {'$ref': '#/definitions/Tooltip'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, field=Undefined, type=Undefined, alt=Undefined, format=Undefined, **kwds):
        super(Tooltip, self).__init__(field=field, type=type, alt=alt, format=format, **kwds)


class Track(GoslingSchema):
    """Track schema wrapper

    anyOf(:class:`SingleTrack`, :class:`OverlaidTrack`, :class:`DataTrack`,
    :class:`TemplateTrack`, :class:`DummyTrack`)
    """
    _schema = {'$ref': '#/definitions/Track'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(Track, self).__init__(*args, **kwds)


class DataTrack(Track):
    """DataTrack schema wrapper

    Mapping(required=[data])
    Partial specification of `BasicSingleTrack` to use default visual encoding predefined by
    data type.

    Attributes
    ----------

    data : :class:`DataDeep`

    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    _invalidTrack : boolean
        internal
    _renderingId : string
        internal
    assembly : :class:`Assembly`
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    endAngle : float
        Specify the end angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    height : float
        Specify the track height in pixels.
    id : string
        Assigned to `uid` in a HiGlass view config, used for API and caching.
    innerRadius : float
        Specify the inner radius of tracks when (`{"layout": "circular"}`).
    layout : :class:`Layout`
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    orientation : :class:`Orientation`
        Specify the orientation.
    outerRadius : float
        Specify the outer radius of tracks when `{"layout": "circular"}`.
    overlayOnPreviousTrack : boolean

    prerelease : Mapping(required=[])
        internal
    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    startAngle : float
        Specify the start angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    style : :class:`Style`
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    subtitle : string

    title : string
        If defined, will show the textual label on the left-top corner of a track.
    width : float
        Specify the track width in pixels.
    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    zoomLimits : :class:`ZoomLimits`

    """
    _schema = {'$ref': '#/definitions/DataTrack'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, data=Undefined, _assignedHeight=Undefined, _assignedWidth=Undefined,
                 _invalidTrack=Undefined, _renderingId=Undefined, assembly=Undefined,
                 centerRadius=Undefined, endAngle=Undefined, height=Undefined, id=Undefined,
                 innerRadius=Undefined, layout=Undefined, linkingId=Undefined, orientation=Undefined,
                 outerRadius=Undefined, overlayOnPreviousTrack=Undefined, prerelease=Undefined,
                 spacing=Undefined, startAngle=Undefined, static=Undefined, style=Undefined,
                 subtitle=Undefined, title=Undefined, width=Undefined, xAxis=Undefined,
                 xDomain=Undefined, xOffset=Undefined, yDomain=Undefined, yOffset=Undefined,
                 zoomLimits=Undefined, **kwds):
        super(DataTrack, self).__init__(data=data, _assignedHeight=_assignedHeight,
                                        _assignedWidth=_assignedWidth, _invalidTrack=_invalidTrack,
                                        _renderingId=_renderingId, assembly=assembly,
                                        centerRadius=centerRadius, endAngle=endAngle, height=height,
                                        id=id, innerRadius=innerRadius, layout=layout,
                                        linkingId=linkingId, orientation=orientation,
                                        outerRadius=outerRadius,
                                        overlayOnPreviousTrack=overlayOnPreviousTrack,
                                        prerelease=prerelease, spacing=spacing, startAngle=startAngle,
                                        static=static, style=style, subtitle=subtitle, title=title,
                                        width=width, xAxis=xAxis, xDomain=xDomain, xOffset=xOffset,
                                        yDomain=yDomain, yOffset=yOffset, zoomLimits=zoomLimits, **kwds)


class DummyTrack(Track):
    """DummyTrack schema wrapper

    Mapping(required=[type])
    A placeholder track. In contrast to other tracks, this track does not display any data.
    Instead it provides empty space for third party tools to display their data on top of.

    Attributes
    ----------

    type : string
        Used to specify the dummy track
    _invalidTrack : boolean
        internal
    assembly : string
        No assemblies can be associated with a dummy track
    height : float
        Specify the track height in pixels.
    id : string
        Assigned to `uid` in a HiGlass view config, used for API and caching.
    layout : string
        Only linear layout are supported at this time
    orientation : :class:`Orientation`
        Specify the orientation.
    overlayOnPreviousTrack : boolean

    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    style : :class:`DummyTrackStyle`
        Defines how the track is styled
    title : string
        Text that gets shown on the DummyTrack
    width : float
        Specify the track width in pixels.
    zoomLimits : List(None)
        Unused property for DummyTrack
    """
    _schema = {'$ref': '#/definitions/DummyTrack'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, _invalidTrack=Undefined, assembly=Undefined, height=Undefined,
                 id=Undefined, layout=Undefined, orientation=Undefined,
                 overlayOnPreviousTrack=Undefined, static=Undefined, style=Undefined, title=Undefined,
                 width=Undefined, zoomLimits=Undefined, **kwds):
        super(DummyTrack, self).__init__(type=type, _invalidTrack=_invalidTrack, assembly=assembly,
                                         height=height, id=id, layout=layout, orientation=orientation,
                                         overlayOnPreviousTrack=overlayOnPreviousTrack, static=static,
                                         style=style, title=title, width=width, zoomLimits=zoomLimits,
                                         **kwds)


class OverlaidTrack(Track):
    """OverlaidTrack schema wrapper

    Mapping(required=[_overlay])
    Superposing multiple tracks.

    Attributes
    ----------

    _overlay : List(Mapping(required=[]))

    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    _invalidTrack : boolean
        internal
    _renderingId : string
        internal
    assembly : :class:`Assembly`
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    baselineY : float

    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    color : anyOf(:class:`Color`, :class:`ChannelValue`)

    data : :class:`DataDeep`

    dataTransform : List(:class:`DataTransform`)

    displacement : :class:`Displacement`

    endAngle : float
        Specify the end angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    experimental : Mapping(required=[])

    flipY : boolean

    height : float
        Specify the track height in pixels.
    id : string
        Assigned to `uid` in a HiGlass view config, used for API and caching.
    innerRadius : float
        Specify the inner radius of tracks when (`{"layout": "circular"}`).
    layout : :class:`Layout`
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    mark : :class:`Mark`

    opacity : anyOf(:class:`Opacity`, :class:`ChannelValue`)

    orientation : :class:`Orientation`
        Specify the orientation.
    outerRadius : float
        Specify the outer radius of tracks when `{"layout": "circular"}`.
    overlayOnPreviousTrack : boolean

    overrideTemplate : boolean

    prerelease : Mapping(required=[])
        internal
    row : anyOf(:class:`Row`, :class:`ChannelValue`)

    size : anyOf(:class:`Size`, :class:`ChannelValue`)

    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    startAngle : float
        Specify the start angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    stretch : boolean

    stroke : anyOf(:class:`Stroke`, :class:`ChannelValue`)

    strokeWidth : anyOf(:class:`StrokeWidth`, :class:`ChannelValue`)

    style : :class:`Style`
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    subtitle : string

    text : anyOf(:class:`Text`, :class:`ChannelValue`)

    title : string
        If defined, will show the textual label on the left-top corner of a track.
    tooltip : List(:class:`Tooltip`)

    visibility : List(:class:`VisibilityCondition`)

    width : float
        Specify the track width in pixels.
    x : anyOf(:class:`X`, :class:`ChannelValue`)

    x1 : anyOf(:class:`X`, :class:`ChannelValue`)

    x1e : anyOf(:class:`X`, :class:`ChannelValue`)

    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    xe : anyOf(:class:`X`, :class:`ChannelValue`)

    y : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1 : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1e : anyOf(:class:`Y`, :class:`ChannelValue`)

    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    ye : anyOf(:class:`Y`, :class:`ChannelValue`)

    zoomLimits : :class:`ZoomLimits`

    """
    _schema = {'$ref': '#/definitions/OverlaidTrack'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, _overlay=Undefined, _assignedHeight=Undefined, _assignedWidth=Undefined,
                 _invalidTrack=Undefined, _renderingId=Undefined, assembly=Undefined,
                 baselineY=Undefined, centerRadius=Undefined, color=Undefined, data=Undefined,
                 dataTransform=Undefined, displacement=Undefined, endAngle=Undefined,
                 experimental=Undefined, flipY=Undefined, height=Undefined, id=Undefined,
                 innerRadius=Undefined, layout=Undefined, linkingId=Undefined, mark=Undefined,
                 opacity=Undefined, orientation=Undefined, outerRadius=Undefined,
                 overlayOnPreviousTrack=Undefined, overrideTemplate=Undefined, prerelease=Undefined,
                 row=Undefined, size=Undefined, spacing=Undefined, startAngle=Undefined,
                 static=Undefined, stretch=Undefined, stroke=Undefined, strokeWidth=Undefined,
                 style=Undefined, subtitle=Undefined, text=Undefined, title=Undefined,
                 tooltip=Undefined, visibility=Undefined, width=Undefined, x=Undefined, x1=Undefined,
                 x1e=Undefined, xAxis=Undefined, xDomain=Undefined, xOffset=Undefined, xe=Undefined,
                 y=Undefined, y1=Undefined, y1e=Undefined, yDomain=Undefined, yOffset=Undefined,
                 ye=Undefined, zoomLimits=Undefined, **kwds):
        super(OverlaidTrack, self).__init__(_overlay=_overlay, _assignedHeight=_assignedHeight,
                                            _assignedWidth=_assignedWidth, _invalidTrack=_invalidTrack,
                                            _renderingId=_renderingId, assembly=assembly,
                                            baselineY=baselineY, centerRadius=centerRadius, color=color,
                                            data=data, dataTransform=dataTransform,
                                            displacement=displacement, endAngle=endAngle,
                                            experimental=experimental, flipY=flipY, height=height,
                                            id=id, innerRadius=innerRadius, layout=layout,
                                            linkingId=linkingId, mark=mark, opacity=opacity,
                                            orientation=orientation, outerRadius=outerRadius,
                                            overlayOnPreviousTrack=overlayOnPreviousTrack,
                                            overrideTemplate=overrideTemplate, prerelease=prerelease,
                                            row=row, size=size, spacing=spacing, startAngle=startAngle,
                                            static=static, stretch=stretch, stroke=stroke,
                                            strokeWidth=strokeWidth, style=style, subtitle=subtitle,
                                            text=text, title=title, tooltip=tooltip,
                                            visibility=visibility, width=width, x=x, x1=x1, x1e=x1e,
                                            xAxis=xAxis, xDomain=xDomain, xOffset=xOffset, xe=xe, y=y,
                                            y1=y1, y1e=y1e, yDomain=yDomain, yOffset=yOffset, ye=ye,
                                            zoomLimits=zoomLimits, **kwds)


class SingleTrack(Track):
    """SingleTrack schema wrapper

    Mapping(required=[data, mark])

    Attributes
    ----------

    data : :class:`DataDeep`

    mark : :class:`Mark`

    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    _invalidTrack : boolean
        internal
    _renderingId : string
        internal
    assembly : :class:`Assembly`
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    baselineY : float

    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    color : anyOf(:class:`Color`, :class:`ChannelValue`)

    dataTransform : List(:class:`DataTransform`)

    displacement : :class:`Displacement`

    endAngle : float
        Specify the end angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    experimental : Mapping(required=[])

    flipY : boolean

    height : float
        Specify the track height in pixels.
    id : string
        Assigned to `uid` in a HiGlass view config, used for API and caching.
    innerRadius : float
        Specify the inner radius of tracks when (`{"layout": "circular"}`).
    layout : :class:`Layout`
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    opacity : anyOf(:class:`Opacity`, :class:`ChannelValue`)

    orientation : :class:`Orientation`
        Specify the orientation.
    outerRadius : float
        Specify the outer radius of tracks when `{"layout": "circular"}`.
    overlayOnPreviousTrack : boolean

    overrideTemplate : boolean

    prerelease : Mapping(required=[])
        internal
    row : anyOf(:class:`Row`, :class:`ChannelValue`)

    size : anyOf(:class:`Size`, :class:`ChannelValue`)

    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    startAngle : float
        Specify the start angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    stretch : boolean

    stroke : anyOf(:class:`Stroke`, :class:`ChannelValue`)

    strokeWidth : anyOf(:class:`StrokeWidth`, :class:`ChannelValue`)

    style : :class:`Style`
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    subtitle : string

    text : anyOf(:class:`Text`, :class:`ChannelValue`)

    title : string
        If defined, will show the textual label on the left-top corner of a track.
    tooltip : List(:class:`Tooltip`)

    visibility : List(:class:`VisibilityCondition`)

    width : float
        Specify the track width in pixels.
    x : anyOf(:class:`X`, :class:`ChannelValue`)

    x1 : anyOf(:class:`X`, :class:`ChannelValue`)

    x1e : anyOf(:class:`X`, :class:`ChannelValue`)

    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    xe : anyOf(:class:`X`, :class:`ChannelValue`)

    y : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1 : anyOf(:class:`Y`, :class:`ChannelValue`)

    y1e : anyOf(:class:`Y`, :class:`ChannelValue`)

    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    ye : anyOf(:class:`Y`, :class:`ChannelValue`)

    zoomLimits : :class:`ZoomLimits`

    """
    _schema = {'$ref': '#/definitions/SingleTrack'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, data=Undefined, mark=Undefined, _assignedHeight=Undefined,
                 _assignedWidth=Undefined, _invalidTrack=Undefined, _renderingId=Undefined,
                 assembly=Undefined, baselineY=Undefined, centerRadius=Undefined, color=Undefined,
                 dataTransform=Undefined, displacement=Undefined, endAngle=Undefined,
                 experimental=Undefined, flipY=Undefined, height=Undefined, id=Undefined,
                 innerRadius=Undefined, layout=Undefined, linkingId=Undefined, opacity=Undefined,
                 orientation=Undefined, outerRadius=Undefined, overlayOnPreviousTrack=Undefined,
                 overrideTemplate=Undefined, prerelease=Undefined, row=Undefined, size=Undefined,
                 spacing=Undefined, startAngle=Undefined, static=Undefined, stretch=Undefined,
                 stroke=Undefined, strokeWidth=Undefined, style=Undefined, subtitle=Undefined,
                 text=Undefined, title=Undefined, tooltip=Undefined, visibility=Undefined,
                 width=Undefined, x=Undefined, x1=Undefined, x1e=Undefined, xAxis=Undefined,
                 xDomain=Undefined, xOffset=Undefined, xe=Undefined, y=Undefined, y1=Undefined,
                 y1e=Undefined, yDomain=Undefined, yOffset=Undefined, ye=Undefined,
                 zoomLimits=Undefined, **kwds):
        super(SingleTrack, self).__init__(data=data, mark=mark, _assignedHeight=_assignedHeight,
                                          _assignedWidth=_assignedWidth, _invalidTrack=_invalidTrack,
                                          _renderingId=_renderingId, assembly=assembly,
                                          baselineY=baselineY, centerRadius=centerRadius, color=color,
                                          dataTransform=dataTransform, displacement=displacement,
                                          endAngle=endAngle, experimental=experimental, flipY=flipY,
                                          height=height, id=id, innerRadius=innerRadius, layout=layout,
                                          linkingId=linkingId, opacity=opacity, orientation=orientation,
                                          outerRadius=outerRadius,
                                          overlayOnPreviousTrack=overlayOnPreviousTrack,
                                          overrideTemplate=overrideTemplate, prerelease=prerelease,
                                          row=row, size=size, spacing=spacing, startAngle=startAngle,
                                          static=static, stretch=stretch, stroke=stroke,
                                          strokeWidth=strokeWidth, style=style, subtitle=subtitle,
                                          text=text, title=title, tooltip=tooltip,
                                          visibility=visibility, width=width, x=x, x1=x1, x1e=x1e,
                                          xAxis=xAxis, xDomain=xDomain, xOffset=xOffset, xe=xe, y=y,
                                          y1=y1, y1e=y1e, yDomain=yDomain, yOffset=yOffset, ye=ye,
                                          zoomLimits=zoomLimits, **kwds)


class TemplateTrack(Track):
    """TemplateTrack schema wrapper

    Mapping(required=[template, data])
    Template specification that will be internally converted into `SingleTrack` for rendering.

    Attributes
    ----------

    data : :class:`DataDeep`

    template : string

    _assignedHeight : float

    _assignedWidth : float
        Internal: Used for responsive spec
    _invalidTrack : boolean
        internal
    _renderingId : string
        internal
    assembly : :class:`Assembly`
        A string that specifies the genome builds to use. Currently support `"hg38"`,
        `"hg19"`, `"hg18"`, `"hg17"`, `"hg16"`, `"mm10"`, `"mm9"`, and `"unknown"`.

        __Note:__: with `"unknown"` assembly, genomic axes do not show chrN: in labels.
    centerRadius : float
        Proportion of the radius of the center white space.

        __Default:__ `0.3`
    encoding : Mapping(required=[])

    endAngle : float
        Specify the end angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    height : float
        Specify the track height in pixels.
    id : string
        Assigned to `uid` in a HiGlass view config, used for API and caching.
    innerRadius : float
        Specify the inner radius of tracks when (`{"layout": "circular"}`).
    layout : :class:`Layout`
        Specify the layout type of all tracks.
    linkingId : string
        Specify an ID for [linking multiple
        views](http://gosling-lang.org/docs/user-interaction#linking-views)
    orientation : :class:`Orientation`
        Specify the orientation.
    outerRadius : float
        Specify the outer radius of tracks when `{"layout": "circular"}`.
    overlayOnPreviousTrack : boolean

    prerelease : Mapping(required=[])
        internal
    spacing : float
        The size of the gap (1) between tracks, (2) between views, and (3) of the origin of
        circular tracks. The effect of this property depends on where on the spec you
        specify the `spacing`.

        In a linear layout, this value is used in pixels, while in a circular layout, this
        value is used relative to the height of the tracks or views.
    startAngle : float
        Specify the start angle (in the range of [0, 360]) of circular tracks (`{"layout":
        "circular"}`).
    static : boolean
        Whether to disable [Zooming and
        Panning](http://gosling-lang.org/docs/user-interaction#zooming-and-panning),
        __Default:__ `false`.
    style : :class:`Style`
        Define the
        [style](http://gosling-lang.org/docs/visual-channel#style-related-properties) of
        multive views. Will be overwritten by the style of children elements (e.g., view,
        track).
    subtitle : string

    title : string
        If defined, will show the textual label on the left-top corner of a track.
    width : float
        Specify the track width in pixels.
    xAxis : :class:`AxisPosition`
        not supported
    xDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic x-axis
    xOffset : float
        Specify the x offset of views in the unit of pixels
    yDomain : anyOf(:class:`DomainInterval`, :class:`DomainChrInterval`, :class:`DomainChr`)
        Specify the visible region of genomic y-axis
    yOffset : float
        Specify the y offset of views in the unit of pixels
    zoomLimits : :class:`ZoomLimits`

    """
    _schema = {'$ref': '#/definitions/TemplateTrack'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, data=Undefined, template=Undefined, _assignedHeight=Undefined,
                 _assignedWidth=Undefined, _invalidTrack=Undefined, _renderingId=Undefined,
                 assembly=Undefined, centerRadius=Undefined, encoding=Undefined, endAngle=Undefined,
                 height=Undefined, id=Undefined, innerRadius=Undefined, layout=Undefined,
                 linkingId=Undefined, orientation=Undefined, outerRadius=Undefined,
                 overlayOnPreviousTrack=Undefined, prerelease=Undefined, spacing=Undefined,
                 startAngle=Undefined, static=Undefined, style=Undefined, subtitle=Undefined,
                 title=Undefined, width=Undefined, xAxis=Undefined, xDomain=Undefined,
                 xOffset=Undefined, yDomain=Undefined, yOffset=Undefined, zoomLimits=Undefined, **kwds):
        super(TemplateTrack, self).__init__(data=data, template=template,
                                            _assignedHeight=_assignedHeight,
                                            _assignedWidth=_assignedWidth, _invalidTrack=_invalidTrack,
                                            _renderingId=_renderingId, assembly=assembly,
                                            centerRadius=centerRadius, encoding=encoding,
                                            endAngle=endAngle, height=height, id=id,
                                            innerRadius=innerRadius, layout=layout, linkingId=linkingId,
                                            orientation=orientation, outerRadius=outerRadius,
                                            overlayOnPreviousTrack=overlayOnPreviousTrack,
                                            prerelease=prerelease, spacing=spacing,
                                            startAngle=startAngle, static=static, style=style,
                                            subtitle=subtitle, title=title, width=width, xAxis=xAxis,
                                            xDomain=xDomain, xOffset=xOffset, yDomain=yDomain,
                                            yOffset=yOffset, zoomLimits=zoomLimits, **kwds)


class ValueExtent(Range):
    """ValueExtent schema wrapper

    anyOf(List(string), List(float))
    """
    _schema = {'$ref': '#/definitions/ValueExtent'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(ValueExtent, self).__init__(*args, **kwds)


class VcfData(DataDeep):
    """VcfData schema wrapper

    Mapping(required=[type, url, indexUrl])
    The Variant Call Format (VCF).

    Attributes
    ----------

    indexUrl : string
        URL link to the tabix index file
    type : string

    url : string
        URL link to the VCF file
    sampleLength : float
        The maximum number of rows to be loaded from the URL. __Default:__ `1000`
    """
    _schema = {'$ref': '#/definitions/VcfData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, indexUrl=Undefined, type=Undefined, url=Undefined, sampleLength=Undefined, **kwds):
        super(VcfData, self).__init__(indexUrl=indexUrl, type=type, url=url, sampleLength=sampleLength,
                                      **kwds)


class VectorData(DataDeep):
    """VectorData schema wrapper

    Mapping(required=[type, url])
    One-dimensional quantitative values along genomic position (e.g., bigwig) can be converted
    into HiGlass' `"vector"` format data. Find out more about this format at [HiGlass
    Docs](https://docs.higlass.io/data_preparation.html#bigwig-files).

    Attributes
    ----------

    type : string

    url : string
        Specify the URL address of the data file.
    aggregation : :class:`BinAggregate`
        Determine aggregation function to apply within bins. __Default__: `"mean"`
    binSize : float
        Binning the genomic interval in tiles (unit size: 256).
    column : string
        Assign a field name of the middle position of genomic intervals. __Default__:
        `"position"`
    end : string
        Assign a field name of the end position of genomic intervals. __Default__: `"end"`
    start : string
        Assign a field name of the start position of genomic intervals. __Default__:
        `"start"`
    value : string
        Assign a field name of quantitative values. __Default__: `"value"`
    """
    _schema = {'$ref': '#/definitions/VectorData'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, type=Undefined, url=Undefined, aggregation=Undefined, binSize=Undefined,
                 column=Undefined, end=Undefined, start=Undefined, value=Undefined, **kwds):
        super(VectorData, self).__init__(type=type, url=url, aggregation=aggregation, binSize=binSize,
                                         column=column, end=end, start=start, value=value, **kwds)


class VisibilityCondition(GoslingSchema):
    """VisibilityCondition schema wrapper

    anyOf(:class:`SizeVisibilityCondition`, :class:`ZoomLevelVisibilityCondition`)
    """
    _schema = {'$ref': '#/definitions/VisibilityCondition'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args, **kwds):
        super(VisibilityCondition, self).__init__(*args, **kwds)


class SizeVisibilityCondition(VisibilityCondition):
    """SizeVisibilityCondition schema wrapper

    Mapping(required=[measure, operation, target, threshold])

    Attributes
    ----------

    measure : enum('width', 'height')
        Specify which aspect of the `target` will be compared to the `threshold`.
    operation : :class:`LogicalOperation`
        A string that specifies the logical operation to conduct between `threshold` and the
        `measure` of `target`. Support

        - greater than : "greater-than", "gt", "GT"

        - less than : "less-than", "lt", "LT"

        - greater than or equal to : "greater-than-or-equal-to", "gtet", "GTET"

        - less than or equal to : "less-than-or-equal-to", "ltet", "LTET"
    target : enum('track', 'mark')
        Target specifies the object that you want to compare with the threshold.
    threshold : anyOf(float, string)
        Specify the threshold as one of:

        - A number representing a fixed threshold in the unit of pixels;

        - `"|xe-x|"`, using the distance between `xe` and `x` as threshold
    conditionPadding : float
        Specify the buffer size (in pixel) of width or height when calculating the
        visibility.

        __Default__: `0`
    transitionPadding : float
        Specify the buffer size (in pixel) of width or height for smooth transition.

        __Default__: `0`
    """
    _schema = {'$ref': '#/definitions/SizeVisibilityCondition'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, measure=Undefined, operation=Undefined, target=Undefined, threshold=Undefined,
                 conditionPadding=Undefined, transitionPadding=Undefined, **kwds):
        super(SizeVisibilityCondition, self).__init__(measure=measure, operation=operation,
                                                      target=target, threshold=threshold,
                                                      conditionPadding=conditionPadding,
                                                      transitionPadding=transitionPadding, **kwds)


class X(ChannelDeep):
    """X schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    domain : :class:`GenomicDomain`
        Values of the data
    field : string
        Name of the data field.
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : string
        Specify the data type.
    """
    _schema = {'$ref': '#/definitions/X'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, aggregate=Undefined, axis=Undefined, domain=Undefined, field=Undefined,
                 grid=Undefined, legend=Undefined, linkingId=Undefined, range=Undefined, type=Undefined,
                 **kwds):
        super(X, self).__init__(aggregate=aggregate, axis=axis, domain=domain, field=field, grid=grid,
                                legend=legend, linkingId=linkingId, range=range, type=type, **kwds)


class Y(ChannelDeep):
    """Y schema wrapper

    Mapping(required=[])

    Attributes
    ----------

    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    baseline : anyOf(string, float)
        Custom baseline of the y-axis. __Default__: `0`
    domain : anyOf(:class:`ValueExtent`, :class:`GenomicDomain`)
        Values of the data
    field : string
        Name of the data field.
    flip : boolean
        Whether to flip the y-axis. This is done by inverting the `range` property.
        __Default__: `false`
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : enum('quantitative', 'nominal', 'genomic')
        Specify the data type.
    zeroBaseline : boolean
        Specify whether to use zero baseline. __Default__: `true`
    """
    _schema = {'$ref': '#/definitions/Y'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, aggregate=Undefined, axis=Undefined, baseline=Undefined, domain=Undefined,
                 field=Undefined, flip=Undefined, grid=Undefined, legend=Undefined, linkingId=Undefined,
                 range=Undefined, type=Undefined, zeroBaseline=Undefined, **kwds):
        super(Y, self).__init__(aggregate=aggregate, axis=axis, baseline=baseline, domain=domain,
                                field=field, flip=flip, grid=grid, legend=legend, linkingId=linkingId,
                                range=range, type=type, zeroBaseline=zeroBaseline, **kwds)


class ZoomLevelVisibilityCondition(VisibilityCondition):
    """ZoomLevelVisibilityCondition schema wrapper

    Mapping(required=[measure, operation, target, threshold])

    Attributes
    ----------

    measure : string
        Specify which aspect of the `target` will be compared to the `threshold`.
    operation : :class:`LogicalOperation`
        A string that specifies the logical operation to conduct between `threshold` and the
        `measure` of `target`. Support

        - greater than : "greater-than", "gt", "GT"

        - less than : "less-than", "lt", "LT"

        - greater than or equal to : "greater-than-or-equal-to", "gtet", "GTET"

        - less than or equal to : "less-than-or-equal-to", "ltet", "LTET"
    target : enum('track', 'mark')
        Target specifies the object that you want to compare with the threshold.
    threshold : float
        Set a threshold in the unit of base pairs (bp)
    conditionPadding : float
        Specify the buffer size (in pixel) of width or height when calculating the
        visibility.

        __Default__: `0`
    transitionPadding : float
        Specify the buffer size (in pixel) of width or height for smooth transition.

        __Default__: `0`
    """
    _schema = {'$ref': '#/definitions/ZoomLevelVisibilityCondition'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, measure=Undefined, operation=Undefined, target=Undefined, threshold=Undefined,
                 conditionPadding=Undefined, transitionPadding=Undefined, **kwds):
        super(ZoomLevelVisibilityCondition, self).__init__(measure=measure, operation=operation,
                                                           target=target, threshold=threshold,
                                                           conditionPadding=conditionPadding,
                                                           transitionPadding=transitionPadding, **kwds)


class ZoomLimits(GoslingSchema):
    """ZoomLimits schema wrapper

    List(anyOf(float, None))
    """
    _schema = {'$ref': '#/definitions/ZoomLimits'}
    _rootschema = GoslingSchema._rootschema

    def __init__(self, *args):
        super(ZoomLimits, self).__init__(*args)




"""
Brushing and Linking
====================
"""
# category: basic marks
import gosling as gos

data = gos.multivec(
    url="https://resgen.io/api/v1/tileset_info/?d=UvVPeLHuRDiYA3qwFlm7xQ",
    row="sample",
    column="position",
    value="peak",
    categories=["sample 1"],
)

base = gos.Track(data).encode(y="peak:Q").properties(width=800, height=200)

top =  gos.overlay(
    base.mark_line().encode(
        x=gos.X("position:G", domain=gos.GenomicDomain(chromosome="chr1"), axis="top")
    ),
    base.mark_brush().encode(
        x=gos.X(
            "position:G",
            domain=gos.GenomicDomain(chromosome="chr1"),
            axis="top",
            linkingId="linking-with-brush"
        ),
        color=gos.value("steelBlue")
    ),
)

bottom = base.mark_line(background="steelBlue", backgroundOpacity=0.1).encode(
    x=gos.X(
        "position:G",
        domain=gos.GenomicDomain(chromosome="chr1", interval=[200000000, 220000000]),
        axis="top",
        linkingId="linking-with-brush",
    )
)

gos.stack(top, bottom)



# derived from https://github.com/altair-viz/altair/blob/8a8642b2e7eeee3b914850a8f7aacd53335302d9/altair/utils/plugin_registry.py
from dataclasses import dataclass
from typing import TypeVar, Any, Generic, Dict, Union, List, cast
import sys
import functools

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

PluginType = TypeVar("PluginType")

@dataclass
class ActivePlugin(Generic[PluginType]):
    name: str
    plugin: PluginType
    options: Dict[str, Any]

class PluginEnabler(Generic[PluginType]):
    """Context manager for enabling plugins.

    This object lets you use enable() as a context manager to
    temporarily enable a given plugin::
        with plugins.enable('name'):
            do_something()  # 'name' plugin temporarily enabled
        # plugins back to original state
    """

    def __init__(
        self,
        registry: "PluginRegistry[PluginType]",
        reset: Union[None, ActivePlugin[PluginType]],
    ):
        self.registry = registry
        self.reset = reset

    def __enter__(self) -> "PluginEnabler[PluginType]":
        return self

    def __exit__(self, *args, **kwargs) -> None:
        self.registry._active = self.reset

    def __repr__(self) -> str:
        return "{}.enable({!r})".format(
            self.registry.__class__.__name__, self.registry.active
        )


class PluginRegistry(Generic[PluginType]):
    """A registry for plugins.

    This is a plugin registry that allows plugins to be loaded/registered
    in two ways:

    1. Through an explicit call to ``.register(name, value)``.
    2. By looking for other Python packages that are installed and provide
       a setuptools entry point group.
    When you create an instance of this class, provide the name of the
    entry point group to use::

        reg = PluginRegister('my_entrypoint_group')

    """

    def __init__(self, entry_point_group: str = ""):
        self.entry_point_group = entry_point_group
        self._active = None  # type: Union[None, ActivePlugin[PluginType]]
        self._plugins = {}  # type: Dict[str, PluginType]

    def register(
        self, name: str, value: Union[PluginType, None]
    ) -> Union[PluginType, None]:
        """Register a plugin by name and value.

        This method is used for explicit registration of a plugin and shouldn't be
        used to manage entry point managed plugins, which are auto-loaded.

        Parameters
        ==========
        name: str
            The name of the plugin.
        value: PluginType or None
            The actual plugin object to register or None to unregister that plugin.

        Returns
        =======
        plugin: PluginType or None
            The plugin that was registered or unregistered.
        """
        if value is None:
            return self._plugins.pop(name, None)
        else:
            self._plugins[name] = value
            return value

    def names(self) -> List[str]:
        """List the names of the registered and entry points plugins."""
        names = list(self._plugins.keys())
        names.extend(e.name for e in entry_points(group=self.entry_point_group))
        return sorted(set(names))

    def enable(self, name: Union[None, str] = None, **options) -> PluginEnabler:
        """Enable a plugin by name.

        This can be either called directly, or used as a context manager.
        Parameters
        ----------
        name : string (optional)
            The name of the plugin to enable. If not specified, then use the
            current active name.
        **options :
            Any additional parameters will be passed to the plugin as keyword
            arguments

        Returns
        -------
        PluginEnabler:
            An object that allows enable() to be used as a context manager
        """
        name = name or self.active

        if name not in self._plugins and (plugin := self._find_plugin(name)):
            self.register(name, plugin)

        prev = self._active
        self._active = ActivePlugin(name, self._plugins[name], options)
        return PluginEnabler(self, reset=prev)

    def _find_plugin(self, name: str) -> Union[PluginType, None]:
        """Find and load named EntryPoint.

        Raises if multiple EntryPoints have conflicting names.

        Parameters
        ==========
        name: str
            The name of the plugin.

        Returns
        =======
        plugin: PluginType or None
            The (un)found plugin
        """
        eps = entry_points(group=self.entry_point_group, name=name)
        if len(eps) == 0:
            return None
        assert len(eps) == 1, f"Conflicting entry-point '{name}'"
        return cast(PluginType, tuple(eps)[0].load())

    @property
    def active(self) -> str:
        """Return the name of the currently active plugin."""
        return "" if self._active is None else self._active.name

    @property
    def options(self) -> Dict[str, Any]:
        """Return the current options dictionary."""
        return {} if self._active is None else self._active.options

    def get(self) -> Union[None, PluginType]:
        """Return the currently active plugin."""
        if active := self._active:
            if active.options and callable(active.plugin):
                return functools.partial(active.plugin, **active.options)
            return active.plugin

    def __repr__(self) -> str:
        return "{}(active={!r}, registered={!r})" "".format(
            self.__class__.__name__, self.active, self.names()
        )



"""
SARS-CoV-2
==========

Data Source: WashU Virus Genome Browser, NCBI, GISAID
"""
# category: others
import gosling as gos

gene_data = gos.csv(
    url="https://s3.amazonaws.com/gosling-lang.org/data/COVID/NC_045512.2-Genes.csv",
    chromosomeField="Accession",
    genomicFields=["Start", "Stop"],
)

genes_base = gos.Track(gene_data).encode(
    x=gos.X("Start:G"),
    xe=gos.Xe("Stop:G"),
).properties(
    width=800,
    height=30,
)


genes = gos.overlay(
    genes_base.mark_rect().encode(
        color=gos.value("#0072B2"),
        stroke=gos.value("white"),
        strokeWidth=gos.value(2),
    ),
    genes_base.mark_rule(linePattern=dict(type="triangleRight", size=10)).encode(
        color=gos.value("white"),
        opacity=gos.value(0.6),
        strokeWidth=gos.value(0),
    ),
    genes_base.mark_text().encode(
        text=gos.Text("Gene symbol:N"),
        color=gos.value("black"),
        stroke=gos.value("white"),
        strokeWidth=gos.value(3),
    ).visibility_le(
        target="mark",
        measure="width",
        threshold="|xe-x|",
        transitionPadding=30,
    ),
).properties(
    title="NC_045512.2 Genes",
)

protein_data = gos.csv(
    url="https://s3.amazonaws.com/gosling-lang.org/data/COVID/sars-cov-2_Sprot_annot_sorted.bed",
    chromosomeField="Accession",
    genomicFields=["Start", "Stop"],
)

domain = [
    "receptor-binding domain (RBD)",
    "receptor-binding motif (RBM)",
    "S1/S2 cleavage site",
    "heptad repeat 1 (HR1)",
    "heptad repeat 2 (HR2)",
]

proteins_base = gos.Track(protein_data).encode(
    x=gos.X("Start:G"),
    row=gos.Row("Protein:N", domain=domain),
).properties(
    width=800,
    height=80,
)

proteins = gos.overlay(
    proteins_base.mark_rect().encode(
        color=gos.Color("Protein:N", domain=domain),
        xe=gos.Xe("Stop:G"),
    ),
    proteins_base.mark_text(textAnchor="end").encode(
        text=gos.Text("Protein:N"),
        color=gos.value("#333"),
        stroke=gos.value("white"),
        strokeWidth=gos.value(3),
    ),
).properties(
    title="S Protein Annotation",
)

recomb_data = gos.csv(
    url="https://s3.amazonaws.com/gosling-lang.org/data/COVID/TRS-L-dependent_recombinationEvents_sorted.bed",
    chromosomeField="Accession",
    genomicFields=["Start1", "Stop1", "Start2", "Stop2"],
    sampleLength=100,
)

recomb = gos.Track(recomb_data).mark_withinLink().encode(
    x=gos.X("Start1:G"),
    xe=gos.Xe("Stop1:G"),
    x1=gos.X1("Start2:G"),
    x1e=gos.X1e("Stop2:G"),
    stroke=gos.value("#0072B2"),
    color=gos.value("#0072B2"),
    opacity=gos.value(0.1),
).properties(
    title="TRS-L-Dependent Recombination Events",
    width=800,
    height=400,
)

gos.stack(genes, proteins, recomb).properties(
    assembly=[["NC_045512.2", 29903]],
    xDomain=gos.DomainInterval([1, 29903]),
)



# flake8: noqa
from .core import *
from .channels import *
SCHEMA_VERSION = 'v0.17.0'
SCHEMA_URL = 'https://raw.githubusercontent.com/gosling-lang/gosling.js/v0.17.0/src/gosling-schema/gosling.schema.json'
THEMES = {'dark', 'ensembl', 'excel', 'ggplot', 'google', 'igv', 'jbrowse', 'light', 'ucsc', 'warm', 'washu'}



"""
Bar Chart
=========
"""
# category: basic marks
import gosling as gos

data = gos.multivec(
    url="https://resgen.io/api/v1/tileset_info/?d=UvVPeLHuRDiYA3qwFlm7xQ",
    row="sample",
    column="position",
    value="peak",
    categories=["sample 1"],
    binSize=5,
)

track = gos.Track(data).mark_bar().encode(
    x="start:G",
    xe="end:G",
    y="peak:Q",
    stroke=gos.value("white"),
    strokeWidth=gos.value(0.5),
).properties(layout="linear", width=725, height=180)

track.view(title="Basic Marks: Bar", subtitle="Tutorial Examples")



"""
Ideograms
=========
"""
# category: composite visualizations
import gosling as gos

def ideogram_with_bars(chromosome: str, width: int):
    """
    Ideograms
    """
    data = gos.csv(
        url="https://raw.githubusercontent.com/sehilyi/gemini-datasets/master/data/UCSC.HG38.Human.CytoBandIdeogram.csv",
        chromosomeField="Chromosome",
        genomicFields=["chromStart", "chromEnd"],
    )

    arms = gos.Track(data).mark_rect().encode(
        color=gos.Color("Stain:N",
            domain=["gneg", "gpos25", "gpos50", "gpos75", "gpos100", "gvar"],
            range=["white", "#D9D9D9", "#979797", "#636363", "black", "#A0A0F2"],
        ),
        x=gos.X("chromStart:G", axis="none"),
        xe="chromEnd:G",
        stroke=gos.value("black"),
        strokeWidth=gos.value(0.5),
    ).transform_filter_not("Stain", oneOf=["acen"])

    arm_labels = arms.mark_text().encode(
        text="Name:N",
        color=gos.Color("Stain:N",
            domain=["gneg", "gpos25", "gpos50", "gpos75", "gpos100", "gvar"],
            range=["black", "#636363", "black", "#D9D9D9", "white", "black"],
        ),
        strokeWidth=gos.value(0)
    ).visibility_lt(
        target='mark',
        measure='width',
        threshold='|xe-x|',
        transitionPadding=10
    )

    centromere_base = gos.Track(data).encode(
        x=gos.X("chromStart:G"),
        xe="chromEnd:G",
        color=gos.value('red'),
    ).transform_filter(
        "Stain", oneOf=["acen"]
    )

    centromere_1 = centromere_base.mark_triangleLeft().transform_filter(
        "Name", include="p"
    )

    centromere_2 = centromere_base.mark_triangleRight().transform_filter(
        "Name", include="q"
    )

    ideogram = gos.overlay(arms, arm_labels, centromere_1, centromere_2).properties(width=width, height=20)

    """
    Stacked Bar Charts
    """
    mvdata = gos.multivec(
        url="https://server.gosling-lang.org/api/v1/tileset_info/?d=cistrome-multivec",
        row="sample",
        column="position",
        value="peak",
        categories=["sample 1", "sample 2"],
        binSize=2,
    )

    bar = gos.Track(mvdata).mark_bar().encode(
        x="start:G",
        xe="end:G",
        y="peak:Q",
        color=gos.Color("sample:N"),
    ).properties(width=width, height=20)

    return gos.stack(bar, ideogram).properties(
        xDomain=gos.GenomicDomain(chromosome=chromosome),
        spacing=0
    )

"""
Composition
"""
gos.parallel(
    ideogram_with_bars("chr1", 1000),
    ideogram_with_bars("chr2", 970),
    ideogram_with_bars("chr3", 800),
    ideogram_with_bars("chr4", 770),
    ideogram_with_bars("chr5", 740)
).properties(
    static=True
)



from __future__ import annotations

import functools
import pathlib
import typing
from dataclasses import dataclass
import hashlib


@dataclass(frozen=True)
class Tileset:
    filepath: pathlib.Path
    tiles_impl: typing.Callable[[typing.Sequence[str]], list[typing.Any]]
    info: typing.Callable[[], typing.Any]
    uid: str
    type: None | str = None

    def tiles(self, tile_ids: typing.Sequence[str]) -> list[typing.Any]:
        return self.tiles_impl(tile_ids)


def create_uid(filepath: pathlib.Path) -> str:
    return hashlib.md5(str(filepath).encode()).hexdigest()[0:8]


def beddb(filepath: pathlib.Path):
    try:
        from clodius.tiles.beddb import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "beddb" data-server.'
        )

    return Tileset(
        filepath=filepath,
        type="beddb",
        tiles_impl=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=create_uid(filepath),
    )


def bigwig(filepath: pathlib.Path):
    try:
        from clodius.tiles.bigwig import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "vector" data-server.'
        )

    return Tileset(
        filepath=filepath,
        type="bigwig",
        tiles_impl=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=create_uid(filepath),
    )


def multivec(filepath: pathlib.Path):
    try:
        from clodius.tiles.multivec import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "multivec" data-server.'
        )

    return Tileset(
        filepath=filepath,
        type="multivec",
        tiles_impl=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=create_uid(filepath),
    )


def cooler(filepath: pathlib.Path):
    try:
        from clodius.tiles.cooler import tiles, tileset_info
    except ImportError:
        raise ImportError(
            'You must have `clodius` installed to use "matrix" data-server.'
        )

    return Tileset(
        filepath=filepath,
        type="cooler",
        tiles_impl=functools.partial(tiles, filepath),
        info=functools.partial(tileset_info, filepath),
        uid=create_uid(filepath),
    )



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

import pandas as pd

from gosling.schema import *
from gosling.api import *
from gosling.data import (
    GoslingDataServer,
    bam,
    bed,
    beddb,
    bigwig,
    csv,
    data_server,
    json,
    matrix,
    multivec,
    vector,
)
from gosling.display import themes, renderers


@pd.api.extensions.register_dataframe_accessor("gos")  # type: ignore
class GosAccessor:
    def __init__(self, df: pd.DataFrame):
        self._df = df

    def csv(self, **kwargs):
        content = self._df.to_csv(index=False) or ""
        url = data_server(content, extension=".csv")
        return dict(type="csv", url=url, **kwargs)



"""
Circos
======
"""
# category: others
import gosling as gos

WIDTH = 725

tileset = gos.vector(
    url="https://resgen.io/api/v1/tileset_info/?d=VLFaiSVjTjW6mkbjRjWREA",
    column="position",
    value="peak",
)

cytoband = gos.csv(
    url="https://raw.githubusercontent.com/sehilyi/gemini-datasets/master/data/UCSC.HG38.Human.CytoBandIdeogram.csv",
    chromosomeField="Chromosome",
    genomicFields=["chromStart", "chromEnd"],
)

segdup = gos.csv(
    url="https://raw.githubusercontent.com/vigsterkr/circos/master/data/5/segdup.txt",
    headerNames=["id", "chr", "p1", "p2"],
    chromosomePrefix="hs",
    chromosomeField="chr",
    genomicFields=["p1", "p2"],
    separator=" ",
    longToWideId="id",
)

bars = gos.Track(tileset).mark_bar().encode(
    x=gos.X("position:G", axis="top"),
    y=gos.Y("peak:Q", axis="right"),
    color=gos.value("#EEEDA1"),
).properties(width=WIDTH, height=60)

ideogram = gos.Track(cytoband).mark_rect().encode(
    color=gos.Color("Stain:N",
        domain=["gneg", "gpos25", "gpos50", "gpos75", "gpos100", "gvar", "acen"],
        range=["white", "#D9D9D9", "#979797", "#636363", "black", "#F0F0F0", "#8D8D8D"],
    ),
    x="chromStart:G",
    xe="chromEnd:G",
    stroke=gos.value("lightgray"),
    strokeWidth=gos.value(0.5),
).properties(width=WIDTH, height=30)


link_base = gos.Track(segdup).mark_withinLink().encode(
    x="p1:G",
    xe="p1_2:G",
    x1="p2:G",
    x1e="P2_2:G",
    opacity=gos.value(0.4)
)

colors = ["#E79F00", "#029F73", "#0072B2", "#CB7AA7", "#D45E00", "#57B4E9", "#EFE441"]

link = gos.overlay(
    link_base
        .encode(stroke=gos.value("lightgray"), strokeWidth=gos.value(1))
        .transform_filter_not("chr", oneOf=["hs1"])
    ,
    link_base
        .encode(stroke=gos.Stroke("chr_2:N", range=colors), strokeWidth=gos.value(1.5))
        .transform_filter("chr", oneOf=["hs1"])
)

gos.stack(bars, ideogram, link).properties(
    title="Circos",
    description="http://circos.ca/intro/genomic_data/",
    layout="circular",
    static=True,
    spacing=1,
    centerRadius=0.3,
)



import json
import pathlib
from typing import Any, Dict
import warnings

import anywidget
import traitlets as t

_esm = pathlib.Path(__file__).parent / "static" / "widget.js"

if not _esm.exists():
    warnings.warn("Widget front-end code not found. Assuming running in CI.")
    _esm = None


class GoslingWidget(anywidget.AnyWidget):
    _esm = _esm
    _viewconf = t.Unicode("null").tag(sync=True)

    location = t.List(t.Union([t.Float(), t.Tuple()]), read_only=True).tag(sync=True)

    def __init__(self, viewconf: Dict[str, Any], **kwargs):
        super().__init__(_viewconf=json.dumps(viewconf), **kwargs)

    def zoom_to(self, view_id: str, pos: str, padding: float = 0, duration: int = 1000):
        msg = json.dumps(["zoomTo", view_id, pos, padding, duration])
        self.send(msg)



import pathlib
from typing import Iterable, TypeVar, Union, Literal

import gosling.display as display
import gosling.utils as utils
from gosling.schema import Undefined, channels, core, mixins

DEFAULT_MARK = "bar"

T = TypeVar("T")

# Aliases & Utils
Data = core.DataDeep


def value(value, **kwargs):
    """Specify a value for use in an encoding"""
    return dict(value=value, **kwargs)


class _EncodingMixin:
    def encode(self: T, *args, **kwargs) -> T:
        """
        Add channel encodings to the track.

        Parameters
        ----------
        column : :class:`Column`

        opacity : :class:`Opacity`

        row : :class:`Row`

        size : :class:`Size`

        stroke : :class:`Stroke`

        strokeWidth : :class:`StrokeWidth`

        text : :class:`Text`

        x : :class:`X`

        x1 : :class:`X1`

        x1e : :class:`X1e`

        xe : :class:`Xe`

        y : :class:`Y`

        y1 : :class:`Y1`

        y1e : :class:`Y1e`

        ye : :class:`Ye`
        """
        # Convert args to kwargs based on their types.
        copy = self.copy()
        kwargs = utils.infer_encoding_types(args, kwargs, channels)
        for key, value in kwargs.items():
            copy[key] = value
        return copy


class _PropertiesMixin:
    def properties(self: T, **kwargs) -> T:
        """Set top-level properties of the View or Track.

        Argument names and types are the same as class initialization.
        """
        copy = self.copy()
        for key, value in kwargs.items():
            setattr(copy, key, value)
        return copy


class _TransformsMixin:
    def _add_transform(self: T, *transforms) -> T:
        copy = self.copy()
        if copy.dataTransform is Undefined:
            copy.dataTransform = []
        copy.dataTransform.extend(transforms)
        return copy

    @utils.use_signature(core.FilterTransform)
    def transform_filter(self: T, field, **kwargs) -> T:
        """Add FilterTransform to Track"""
        return self._add_transform(
            core.FilterTransform(type="filter", field=field, **kwargs)
        )

    @utils.use_signature(core.FilterTransform)
    def transform_filter_not(self: T, field, **kwargs) -> T:
        """Add FilterTransform with `not=True` to Track"""
        kwargs["not"] = True
        return self._add_transform(
            core.FilterTransform(type="filter", field=field, **kwargs)
        )

    @utils.use_signature(core.LogTransform)
    def transform_log(self: T, field, **kwargs) -> T:
        """Add LogTransform to Track"""
        return self._add_transform(
            core.LogTransform(type="log", field=field, **kwargs),
        )

    @utils.use_signature(core.StrConcatTransform)
    def transform_str_concat(self: T, fields, **kwargs) -> T:
        """Add StrConcatTransform to Track"""
        return self._add_transform(
            core.StrConcatTransform(type="concat", fields=fields, **kwargs),
        )

    @utils.use_signature(core.StrReplaceTransform)
    def transform_str_replace(self: T, field, **kwargs) -> T:
        """Add StrReplaceTransform to Track"""
        return self._add_transform(
            core.StrReplaceTransform(type="replace", field=field, **kwargs)
        )

    @utils.use_signature(core.DisplaceTransform)
    def transform_displace(self: T, **kwargs) -> T:
        """Add DisplaceTransform to Track"""
        return self._add_transform(core.DisplaceTransform(type="displace", **kwargs))

    @utils.use_signature(core.ExonSplitTransform)
    def transform_exon_split(self: T, **kwargs) -> T:
        """Add ExonSplitTransform to Track"""
        return self._add_transform(core.ExonSplitTransform(type="exonSplit", **kwargs))

    @utils.use_signature(core.CoverageTransform)
    def transform_coverage(self: T, startField, endField, **kwargs) -> T:
        """Add CoverageTransform to Track"""
        return self._add_transform(
            core.CoverageTransform(
                type="coverage", startField=startField, endField=endField, **kwargs
            )
        )

    @utils.use_signature(core.JsonParseTransform)
    def transform_json_parse(self: T, field, **kwargs) -> T:
        """Add JSONParseTransform to Track"""
        return self._add_transform(
            core.JsonParseTransform(type="subjson", field=field, **kwargs)
        )


class _VisibilityMixin:
    def _add_visibility(self: T, *visibilities) -> T:
        copy = self.copy()
        if copy.visibility is Undefined:
            copy.visibility = []
        copy.visibility.extend(visibilities)
        return copy

    @utils.use_signature(core.VisibilityCondition)
    def visibility_lt(self: T, **kwargs) -> T:
        """Add less-than VisibilityCondition to Track"""
        return self._add_visibility(core.VisibilityCondition(operation="LT", **kwargs))

    @utils.use_signature(core.VisibilityCondition)
    def visibility_gt(self: T, **kwargs) -> T:
        """Add greater-than VisibilityCondition to Track"""
        return self._add_visibility(core.VisibilityCondition(operation="GT", **kwargs))

    @utils.use_signature(core.VisibilityCondition)
    def visibility_le(self: T, **kwargs) -> T:
        """Add less-than-or-equal-to VisibilityCondition to Track"""
        return self._add_visibility(
            core.VisibilityCondition(operation="LTET", **kwargs)
        )

    @utils.use_signature(core.VisibilityCondition)
    def visibility_ge(self: T, **kwargs) -> T:
        """Add greater-than-or-equal-to VisibilityCondition to Track"""
        return self._add_visibility(
            core.VisibilityCondition(operation="GTET", **kwargs)
        )


@utils.use_signature(core.Root)
class View(_PropertiesMixin, core.Root):
    """Create a basic Gosling View."""

    def _repr_mimebundle_(self, include=None, exclude=None):
        render = display.renderers.get()
        return render(self.to_dict())

    def display(self):
        """Render top-level chart using IPython.display."""
        from IPython.display import display

        display(self)

    def save(
        self,
        path: Union[pathlib.Path, str],
        format: Union[Literal["json", "html"], None] = None,
        **kwargs,
    ):
        if format is None:
            suffix = pathlib.Path(path).suffix.lstrip(".")
            if suffix in ("json", "html"):
                format = suffix
            else:
                raise ValueError("must specify file format: ['html', 'json']")

        if format == "json":
            data = self.to_json(**kwargs)
        elif format == "html":
            data = display.spec_to_html(self.to_dict(), **kwargs)
        else:
            raise ValueError(f"unrecognized format: '{format}'")

        with open(path, mode="w") as f:
            f.write(data)

    def widget(self):
        try:
            from ._widget import GoslingWidget
        except ImportError:
            raise ImportError(
                "The 'gosling-widget' package is required to use the widget() method."
            )

        return GoslingWidget(self.to_dict())


# View utilities


class Track(
    _EncodingMixin,
    _PropertiesMixin,
    _TransformsMixin,
    _VisibilityMixin,
    mixins.MarkMethodMixin,
    core.SingleTrack,
):
    """Create a basic Gosling Track.

    Although it is possible to set all Track properties as constructor attributes,
    it is more idiomatic to use methods such as ``mark_point()``, ``encode()``,
    ``transform_filter()``, ``properties()``, etc.
    """

    @utils.use_signature(core.SingleTrack)
    def __init__(self, data=Undefined, **kwargs):
        super().__init__(
            data=data,
            mark=kwargs.pop("mark", DEFAULT_MARK),
            **kwargs,
        )

    @utils.use_signature(core.SingleView)
    def view(self, **kwargs) -> View:
        """Convert Track into top-level gos.View"""
        return View(tracks=[self], **kwargs)


def overlay(*tracks: Union[Track, View], **kwargs) -> View:
    """Compose an overlaid view from multiple tracks or overliad tracks"""
    return View(alignment="overlay", tracks=tracks, **kwargs)


def stack(*tracks: Union[Track, View], **kwargs) -> View:
    """Compose a stacked view from multiple tracks or overlaid tracks."""
    return View(alignment="stack", tracks=tracks, **kwargs)


def _auto_cast_to_view(views: Iterable[Union[View, Track]]):
    """Converts a list of Tracks and/or Views into a list of Views."""
    return tuple(v.view() if isinstance(v, Track) else v for v in views)


def horizontal(*views: Union[View, Track], **kwargs) -> View:
    """Arrange multiple views horizontally."""
    views = _auto_cast_to_view(views)
    return View(arrangement="horizontal", views=views, **kwargs)


def vertical(*views: Union[View, Track], **kwargs) -> View:
    """Arrange multiple views vertically."""
    views = _auto_cast_to_view(views)
    return View(arrangement="vertical", views=views, **kwargs)


def parallel(*views: Union[View, Track], **kwargs) -> View:
    """Arrange multiple views in parallel."""
    views = _auto_cast_to_view(views)
    return View(arrangement="parallel", views=views, **kwargs)


def serial(*views: Union[View, Track], **kwargs) -> View:
    """Arrange multiple views in serial."""
    views = _auto_cast_to_view(views)
    return View(arrangement="serial", views=views, **kwargs)



"""
Single-cell Epigenomic Analysis
===============================

`Corces et al. 2020`_

.. _Corces et al. 2020: https://www.nature.com/articles/s41588-020-00721-x
"""
# category: others
import gosling as gos

WIDTH = 600

# DATA

cytogenetic_band = gos.csv(
    url="https://raw.githubusercontent.com/sehilyi/gemini-datasets/master/data/cytogenetic_band.csv",
    chromosomeField="Chr.",
    genomicFields=["ISCN_start", "ISCN_stop", "Basepair_start", "Basepair_stop"],
)

gene_anno = gos.beddb(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=gene-annotation",
    genomicFields=[{"index": 1, "name": "start"}, {"index": 2, "name": "end"}],
    valueFields=[
        {"index": 5, "name": "strand", "type": "nominal"},
        {"index": 3, "name": "name", "type": "nominal"}
    ],
    exonIntervalFields=[{"index": 12, "name": "start"}, {"index": 13, "name": "end"}],
)

# TRACKS

ideo_base = gos.Track(cytogenetic_band).mark_rect(outlineWidth=1).encode(
    x=gos.X("Basepair_start:G", axis="none"),
    xe="Basepair_stop:G",
    stroke=gos.value("black"),
    strokeWidth=gos.value(1),
)

ideogram = gos.overlay(
    ideo_base.encode(
        color=gos.Color("Density:N",
            domain=["", "25", "50", "75", "100"],
            range=["white", "#D9D9D9", "#979797", "#636363", "black"]
        ),
        size=gos.value(20),
    ).transform_filter_not("Stain", oneOf=["acen-1", "acen-2"]),
    ideo_base.encode(
        color=gos.value("#A0A0F2"),
        size=gos.value(20),
    ).transform_filter("Stain", oneOf=["gvar"]),
    ideo_base.mark_triangleRight().encode(
        color=gos.value("#B40101"),
        size=gos.value(20),
    ).transform_filter("Stain", oneOf=["acen-1"]),
    ideo_base.mark_triangleLeft().encode(
        color=gos.value("#B40101"),
        size=gos.value(20)
    ).transform_filter("Stain", oneOf=["acen-2"]),
    ideo_base.mark_brush().encode(
        x={"linkingId": "detail"},
        color=gos.value("red"),
        opacity=gos.value(1),
        strokeWidth=gos.value(1),
        stroke=gos.value("red"),
    )
).properties(title="chr3", width=WIDTH, height=25)

gene_base = gos.Track(gene_anno).encode(
    color=gos.Color("strand:N", domain=["+", "-"], range=["#012DB8", "#BE1E2C"]),
    row=gos.Row("strand:N", domain=["+", "-"]),
).visibility_lt(
    measure="width",
    threshold="|xe-x|",
    transitionPadding=10,
    target="mark",
)
genes = gos.overlay(
    gene_base
        .mark_text(textFontSize=8, dy=-12, outline="#20102F")
        .encode(text="name:N", x="start:G", xe="end:G", size=gos.value(8))
        .transform_filter("type", oneOf=["gene"])
        .transform_filter("strand", oneOf=["+"]),
    gene_base
        .mark_text(textFontSize=8, dy=10, outline="#20102F")
        .encode(text="name:N", x="start:G", xe="end:G", size=gos.value(8))
        .transform_filter("type", oneOf=["gene"])
        .transform_filter("strand", oneOf=["-"]),
    gene_base
        .mark_rect(outline="#20102F")
        .encode(x="end:G", size=gos.value(7))
        .transform_filter("type", oneOf=["gene"])
        .transform_filter("strand", oneOf=["+"]),
    gene_base
        .mark_rect(outline="#20102F")
        .encode(x="start:G", size=gos.value(7))
        .transform_filter("type", oneOf=["gene"])
        .transform_filter("strand", oneOf=["-"]),
    gene_base
        .mark_rect(outline="#20102F")
        .encode(x="start:G", xe="end:G", size=gos.value(14))
        .transform_filter("type", oneOf=["exon"]),
    gene_base
        .mark_rule(outline="#20102F")
        .encode(x="start:G", xe="end:G", strokeWidth=gos.value(3))
        .transform_filter("type", oneOf=["exon"]),
)

def bar(title: str, color: str, file: str) -> gos.Track:
    url = f"https://s3.amazonaws.com/gosling-lang.org/data/{file}"
    data = gos.bigwig(url, column="position", value="peak")
    return gos.Track(data).mark_bar(outline="#20102F").encode(
        x="position:G",
        y=gos.Y("peak:Q", axis="right"),
        color=gos.value(color),
    ).properties(title=title, width=WIDTH, height=40)

def link(tileset_id: str, stroke: str) -> gos.Track:
    url = f"https://server.gosling-lang.org/api/v1/tileset_info/?d={tileset_id}"
    data = gos.beddb(url, genomicFields=[{"name": "start", "index": 1}, {"name": "end", "index": 2}])
    return gos.Track(data).mark_withinLink().encode(
        x="start:G",
        xe="end:G",
        y={"flip": True},
        strokeWidth=gos.value(1),
        color=gos.value("none"),
        stroke=gos.value(stroke),
        opacity=gos.value(0.1),
    )

tracks = [
    bar("Excitatory neurons", "#F29B67", "ExcitatoryNeurons-insertions_bin100_RIPnorm.bw"),
    bar("Inhibitory neurons", "#3DC491", "InhibitoryNeurons-insertions_bin100_RIPnorm.bw"),
    bar("Dopaminergic neurons", "#565C8B", "DopaNeurons_Cluster10_AllFrags_projSUNI2_insertions_bin100_RIPnorm.bw"),
    bar("Microglia", "#77C0FA", "Microglia-insertions_bin100_RIPnorm.bw"),
    bar("Oligodendrocytes", "#9B46E5", "Oligodendrocytes-insertions_bin100_RIPnorm.bw"),
    bar("Astrocytes", "#D73636", "Astrocytes-insertions_bin100_RIPnorm.bw"),
    bar("OPCs", "#E38ADC", "OPCs-insertions_bin100_RIPnorm.bw"),
    genes.properties(title="Genes", width=WIDTH, height=80),
    gos.overlay(
        link("oligodendrocyte-plac-seq-bedpe", "#F97E2A"),
        link("microglia-plac-seq-bedpe", "#50ADF9"),
        link("neuron-plac-seq-bedpe", "#7B0EDC"),
    ).properties(title="PLAC-seq (H3K4me3) Nott et al.", width=WIDTH, height=60)
]

# COMPOSE

gos.vertical(
    gos.View(layout="linear", tracks=[ideogram], centerRadius=0.8, xDomain=gos.GenomicDomain(chromosome="chr3")),
    gos.View(tracks=tracks, linkingId="detail", xDomain=gos.GenomicDomain(chromosome="chr3", interval=[52168000, 52890000])),
)



"""
Overview + Details
==================
"""
# category: interactive visualizations
import gosling as gos

multivec_data = gos.multivec(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=cistrome-multivec",
    row="sample",
    column="position",
    value="peak",
    categories=["sample 1", "sample 2", "sample 3", "sample 4"],
    binSize=4,
)

rearrangments = gos.csv(
    url="https://raw.githubusercontent.com/sehilyi/gemini-datasets/master/data/rearrangements.bulk.1639.simple.filtered.pub",
    headerNames=["chr1", "p1s", "p1e", "chr2", "p2s", "p2e", "type", "id", "f1", "f2", "f3", "f4", "f5", "f6"],
    separator="\t",
    genomicFieldsToConvert=[
        {"chromosomeField": "chr1", "genomicFields": ["p1s", "p1e"]},
        {"chromosomeField": "chr2", "genomicFields": ["p2s", "p2e"]},
    ],
)

base = gos.Track(multivec_data).mark_bar(outlineWidth=0).encode(
    x="start:G",
    xe="end:G",
    y="peak:Q",
    row="sample:N",
    color="sample:N",
    stroke=gos.value("black"),
    strokeWidth=gos.value(0.3),
)

circular = gos.stack(
    (
        gos.overlay(
            base.mark_bar(),
            base.mark_brush().encode(
                x=gos.X("start:G", linkingId="detail-1"),
                color=gos.value("blue")
            ),
            base.mark_brush().encode(
                x=gos.X("start:G", linkingId="detail-2"),
                color=gos.value("red")
            ),
        ).properties(width=500, height=100)
    ),
    (
        gos.Track(rearrangments)
            .mark_withinLink()
            .encode(
                x="p1s:G",
                xe="p1e:G",
                x1="p2s:G",
                x1e="p2e:G",
                stroke=gos.Stroke(
                    "type:N",
                    domain=["deletion", "inversion", "translocation", "tandem-duplication"],
                ),
                strokeWidth=gos.value(0.8),
                opacity=gos.value(0.15),
            )
            .transform_filter("chr1", oneOf=["1", "16", "14", "9", "6", "5", "3"])
            .transform_filter("chr2", oneOf=["1", "16", "14", "9", "6", "5", "3"])
            .properties(width=500, height=100)
    ),
).properties(static=True, layout="circular")


def detail(background, linkingId, legend, chromosome):
    domain = gos.GenomicDomain(chromosome=chromosome)
    return base.mark_bar(background=background, backgroundOpacity=0.1).encode(
        x=gos.X("start:G", linkingId=linkingId, domain=domain),
        strokeWidth=gos.value(0.3),
        color=gos.Color("sample:N", legend=legend),
    ).properties(width=245, height=150)

details = gos.horizontal(
    detail("blue", "detail-1", legend=False, chromosome="chr5"),
    detail("red", "detail-2", legend=True, chromosome="chr16"),
    spacing=10,
)

gos.vertical(circular, details)
